from .setup_tests import *

from exoatlas import *


def test_load_references():
    e = Exoplanets()
    e.load_individual_references()
    p = e["GJ1132b"]
    assert len(p.individual_references.standard) > len(p.standard)


def test_check_and_update_references():
    e = Exoplanets()
    e.load_individual_references()
    e.display_individual_references(planets="HD189733b")
    e.update_reference(references="Ivshina + Winn 2022")
    assert (
        e["HD189733b"].get_values_from_table("period_reference")[0]
        == "Ivshina + Winn 2022"
    )


def test_update_values():
    e = Exoplanets()
    e.update_values(
        planets="GJ1214b", mass=1 * u.M_earth, mass_uncertainty=0.1 * u.M_earth
    )
    one = e["GJ1214b"]
    assert one.get("mass") == 1 * u.M_earth
    assert one.get_uncertainty("mass") == 0.1 * u.M_earth
    assert one.get_uncertainty_lowerupper("mass")[0] == 0.1 * u.M_earth



from .setup_tests import *

from exoatlas import *


def test_exoplanets_downloader():
    planet_name = "GJ 1132 b"

    for t in ["ps", "pscomppars"]:
        fresh = ExoplanetArchiveDownloader(which_table=t, planet=planet_name).get_table(
            remake=True
        )
        loaded = ExoplanetArchiveDownloader(
            which_table=t, planet=planet_name
        ).get_table(remake=False)
        assert np.all(loaded["ra"] == fresh["ra"])



"""
A simple cartoon of the Chen & Kipping (2017) forecaster model,
using the values quoted in their Table 2.
"""

__all__ = [
    "use_chen_and_kipping_to_estimate_radius_from_mass",
    "use_chen_and_kipping_to_estimate_mass_from_radius",
    "plot_chen",
]

import numpy as np
import matplotlib.pyplot as plt
from astropy.units import Rearth, Mearth, Rjupiter, Mjupiter, Rsun, Msun

# using log polynomials from https://exoplanetarchive.ipac.caltech.edu/docs/pscp_calc.html


def use_chen_and_kipping_to_estimate_radius_from_mass(M):
    """
    Estimate the radii, given an array of masses.

    Parameters
    ----------
    M : astropy.units.quantity.quantity
        The masses of the planets.
    """

    ### RADIUS FROM MASS!
    M_min, M_max, C, S = {}, {}, {}, {}

    # define the terran range
    M_min["terran"] = 0 * Mearth
    M_max["terran"] = 2.04 * Mearth
    S["terran"] = 0.2790
    C["terran"] = 0.00346

    # define the neptunian range
    M_min["neptunian"] = M_max["terran"]
    M_max["neptunian"] = 132 * Mearth
    S["neptunian"] = 0.589
    C["neptunian"] = -0.0925

    # define the jovian range
    M_min["jovian"] = M_max["neptunian"]
    M_max["jovian"] = 26600 * Mearth
    S["jovian"] = -0.044
    C["jovian"] = 1.25

    # define the jovian range
    M_min["stellar"] = M_max["jovian"]
    M_max["stellar"] = 1 * Msun
    S["stellar"] = 0.881
    C["stellar"] = -2.85

    # create an empty array of radii
    original_shape = np.shape(M)
    logM = np.atleast_1d(np.log10(M.to_value(Mearth)))
    logR = np.zeros(np.shape(logM)) * np.nan

    # work our way down in mass, updating radii as we go
    for this in ["stellar", "jovian", "neptunian", "terran"]:
        # which planets are below this maximum?
        x = M < M_max[this]

        # calculate the radius
        logR[x] = C[this] + logM[x] * S[this]

    # return the same shape
    return 10 ** logR.reshape(original_shape) * Rearth


def use_chen_and_kipping_to_estimate_mass_from_radius(R):
    """
    Estimate the mass, given an array of raii.

    Parameters
    ----------
    R : astropy.units.quantity.quantity
        The radii of the planets.
    """

    ### MASS FROM RADIUS!
    R_min, R_max, C, S = {}, {}, {}, {}

    # define the terran range
    R_min["terran"] = 0 * Rearth
    R_max["terran"] = 1.23 * Rearth
    S["terran"] = 0.2790
    C["terran"] = 0.00346

    # define the neptunian range
    R_min["neptunian"] = R_max["terran"]
    R_max["neptunian"] = 11.1 * Rearth
    S["neptunian"] = 0.589
    C["neptunian"] = -0.0925

    # create an empty array of radii
    original_shape = np.shape(R)
    logR = np.atleast_1d(np.log10(R.to_value(Rearth)))
    logM = np.zeros(np.shape(logR)) * np.nan

    # work our way down in mass, updating radii as we go
    for this in ["neptunian", "terran"]:
        # which planets are below this maximum?
        x = R < R_max[this]

        # calculate the radius
        logM[x] = (logR[x] - C[this]) / S[this]

    # return the same shape
    return 10 ** logM.reshape(original_shape) * Mearth


def plot_chen(independent="mass", **kw):
    """
    Plot the MAP mass-radius curve from Chen and Kipping (2017).
    This is mostly a wrapper to make testing easier.
    """

    if independent[0].lower() == "m":
        M = np.logspace(-1, 5, 1000) * Mearth
        R = use_chen_and_kipping_to_estimate_radius_from_mass(M)
        plt.loglog(M, R, **kw)
    elif independent[0].lower() == "r":
        R = np.logspace(-1, 5, 1000) * Rearth
        M = use_chen_and_kipping_to_estimate_mass_from_radius(R)
        plt.loglog(M, R, **kw)


"""def update_masses_with_chen_and_kipping(population, planets):
    # KLUDGE!?!?!
    # MODIFIES POPULATION IN PLACE!
    population.standard.loc
    mass_is_bad = np.isfinite(population.mass) == False
    population.mass[mass_is_bad] = use_chen_and_kipping_to_estimate_mass_from_radius(population.radius[mass_is_bad])
"""



from ...imports import *
from ..predefined import PredefinedPopulation
from .planets import SolarSystem

__all__ = ["SolarSystemMoons"]

initial_filename = os.path.join(
    code_directory, "populations/data/solarsystem/jpl-ssd-moons.txt"
)


class SolarSystemMoons(PredefinedPopulation):
    """
    The Solar System Moons, very crudely.
    All orbital parameters are of the planet around its star.
    """

    label = "Solar System Moons"

    # the data in the table probably don't need to be updated
    _expiration = np.inf * u.day

    def __init__(self, **kwargs):
        """
        Initialize a population of Solar System (main) planets.
        """
        PredefinedPopulation.__init__(self, **kwargs)
        self.color = "mediumblue"
        self.zorder = None
        self.s = 20
        self.respond_to_color = False
        self.exact = True
        self.marker = "s"

    def _download_raw_data(self, **kw):
        """
        Load the raw table of data for the Solar System.
        """

        # load the table of Solar System planets
        raw = ascii.read(initial_filename)

        # for debugging, hang on to the raw table as a hidden attribute
        self._raw = raw

        # a table of unstandardized planet properties
        return raw

    def _create_standardized(self, raw):
        """
        Create a standardized table of planet properties.
        """

        # start from the raw trimmed table
        t = raw

        s = QTable()
        s["name"] = t["satellite"]
        s["planet"] = t["planet"]
        s["code"] = t["code"]
        s["mass"] = (t["GM"] * u.km**3 / u.s**2 / con.G).to(u.Mearth)
        s["radius"] = (t["radius"] * u.km).to(u.Rearth)
        s["density"] = t["density"] * u.g / u.cm**3
        for k in ["lower", "upper"]:
            s[f"mass_uncertainty_{k}"] = (
                t["GM_uncertainty"] * u.km**3 / u.s**2 / con.G
            ).to(u.Mearth)
            s[f"radius_uncertainty_{k}"] = (t["radius_uncertainty"] * u.km).to(u.Rearth)
            s[f"density_uncertainty_{k}"] = t["density_uncertainty"] * u.g / u.cm**3

        # get stellar properties from the Solar System
        solar = SolarSystem()
        for k in ["mass", "radius", "teff", "age"]:
            s[f"stellar_{k}"] = solar.get(f"stellar_{k}")[0]
        s["hostname"] = "Sun"

        # get orbit properties from the Solar System planet hosts
        planets = solar.standard["name", "period", "semimajoraxis"]
        planets.columns[0].name = "planet"

        merged = join(planets, s, keys="planet")

        return merged



# general class for exoplanet populations
from ..imports import *

# from ..telescopes import *
# from ..models import *
from .column_descriptions import *
from .pineda_skew import make_skew_samples_from_lowerupper, gaussian_central_1sigma

# these are keywords that can be set for a population
default_plotkw = dict(
    color="black",
    alpha=1,
    zorder=0,
    marker="o",
    linewidth=1,
    respond_to_color=True,
    respond_to_size=True,
    exact=False,
    annotate_planets=False,
    filled=True,
    outlined=False,
)

# what keywords can we set for the population plotkw?
allowed_plotkw = list(default_plotkw.keys())
allowed_plotkw += [
    "s",
    "size",
    "c",
    "cmap",
    "norm",
    "vmin",
    "vmax" "outlined",
    "filled",
    "markeredgewidth",
    "markeredgecolor",
]


def _clean_column(raw_column):
    """
    Convert a Column into an array disconnected from its original Table.

    Parameters
    ----------
    raw_column : astropy.table.column.Column
        The column to clean.

    Returns
    -------
    cleaned_column : array, Quantity, Time, SkyCoord, ...
        The data as an array-like object, without
        connection to its original table.
    """
    # strip the connection to the table column
    if isinstance(raw_column, Column):
        cleaned_column = np.array(raw_column)
    else:
        cleaned_column = type(raw_column)(raw_column)
        # This is equivalent to
        #
        # if isinstance(raw_column, u.Quantity):
        #   cleaned_column = u.Quantity(raw_column)
        #
        # I'm not 100% sure why this is necessary, but
        # in some tests it seems a QTable will return a
        # Quantity or SkyCoord or Time object, but still
        # with some sort of connection to being a Column
        # from a Table (and therefore potentially some
        # sneaky link that we want to erase). Re-initializing
        # an object seems to erase that connection.
    return cleaned_column


class Population:
    """
    Populations of astronomical objects might contain
        Exoplanets (planets with host stars),
        Solar System objects (major/minor planets),
        Stars (single or multiples),
    or more!
    """

    def __init__(self, standard, label=None, **plotkw):
        """
        Initialize a Population of exoplanets from a standardized table.

        This creates a population from a standardized data table,
        effectively just storing a table in the right format inside
        the object. To simplify indexing via system name without having
        to stress about capitalization and/or spaces, the population
        will define a `tidyname` and `tidyhostname` columns, and use
        those as table indices for searching via name.

        Parameters
        ----------
        standard : astropy.table.QTable, str
            If a Table, the standardized data table.
            If a string, the filename to the standardized data table.
        label : str
            A string by which this table can be identified. This will be
            used for plots and for saving data files, so please try to
            pick something informative and unique!
        **plotkw : dict
            All other keyword arguments will be taken as plotting suggestions.
        """

        if isinstance(standard, Table) or isinstance(standard, Row):
            # a standardized table with a minimum set of columns we can expect
            self.standard = QTable(copy.deepcopy(standard))
            # (the deepcopy seems to be needed to reset indexing,
            #  so that astropy.table doesn't try to track slices across
            #  too many subsets, because it was getting lost in weird ways)

            # store a label for this population
            self.label = label

            # keywords to use for plotting
            self._plotkw = plotkw

        elif isinstance(standard, str):
            filename = standard
            self.standard = QTable(ascii.read(filename))
            self.label = self.standard.meta["label"]
            self._plotkw = self.standard.meta["plotkw"]

        # define some cleaned names and hostnames, for indexing
        try:
            self.standard["tidyname"]
        except KeyError:
            self.standard["tidyname"] = [
                clean(x).lower() for x in self.standard["name"]
            ]
        try:
            self.standard["tidyhostname"]
        except KeyError:
            self.standard["tidyhostname"] = [
                clean(x).lower() for x in self.standard["hostname"]
            ]

        # make sure the table is searchable via names
        self._make_sure_index_exists("tidyname")
        self._make_sure_index_exists("tidyhostname")

        # test that indexing still works
        name = self.standard["tidyname"][0]
        self.standard.loc[name]

        # define internal lists of column names
        self._populate_column_summaries()
        self._populate_column_methods()

        # define how many samples and iterations to use for uncertainty propagation
        self.targeted_fractional_uncertainty_precision = 0.05
        self._number_of_uncertainty_samples = 100

    def _create_function_to_access_table_quantity(self, name):
        """
        For a given column name "x", add a `.x()` method to this Population.

        This internal wrapper creates a named method to access
        the data in a particular column of the standardized table.
        It's a little frivolous, but this mostly makes it easier to
        see variables via tab completion and for doing error propagation.
        This will be called at the initial creation of the Population,
        and whenever new columns are added via `.add_column`.

        Parameters
        ----------
        name : str
            The name of the column to add.
        """

        assert name in self.standard.colnames

        # create the method to extract a particular column
        def f(distribution=False, **kw):
            return self.get_values_from_table(name, distribution=distribution)

        # build a basic docstring for that method.
        try:
            unit = self.get_values_from_table(name).unit.to_string()
            unit_string = f", with units of '{unit}'"
        except AttributeError:
            unit_string = ""

        f.__doc__ = f"""
            The table quantity '{name}'{unit_string}.

            Parameters 
            ----------
            distribution : bool 
                Should we try to return an astropy Distribution, 
                for uncertainty propagation? If the table contains 
                quoted uncertainties, a distribution will be returned; 
                if not, a simple array will be.
            kw : dict 
                All other keyword arguments will be ignored.

            Returns 
            -------
            values : array-like, or Distribution
                The values for this table column. 
                If `distribution==True`, these values will be as an 
                astropy Distribution, built from the table uncertainties.
            """
        return f

    def _populate_column_methods(self):
        """
        Populate this object with one method for each table column.

        This wrapper converts table columns from the internal .standard
        table into callable methods, so that (for example), the data
        in `self.standard['radius']` can be retrieved as `self.radius()`.
        This is necessary for seamlesssly integrating core table quantities
        with derived quantities (some of which must be callable because
        they require keyword arguments), for making them appear as hints
        with tab-completion in ipython/jupyter, and for being able to
        attach a docstring to each.
        """

        # add all the basic table quantities as methods to the population
        for k in self._colnames["everything"]:
            # if no other function is defined, populate a table one
            if hasattr(self, k) == False:
                setattr(self, k, self._create_function_to_access_table_quantity(k))

    def _populate_column_summaries(self):
        """
        Populate a dictionary summarizing the types of columns in `.standard`.
        """
        uncertainty_suffixes = [
            "_uncertainty_lower",
            "_uncertainty_upper",
            "_uncertainty",
        ]
        limit_suffixes = ["_lower_limit", "_upper_limit"]
        reference_suffixes = ["_reference"]
        all_suffixes = uncertainty_suffixes + limit_suffixes + reference_suffixes

        def ends_with_any_of(s, suffixes):
            return np.any([s.endswith(suffix) for suffix in suffixes])

        def remove_suffixes(s):
            for suffix in all_suffixes:
                s = s.split(suffix)[0]
            return s

        self._colnames = {}
        self._colnames["everything"] = np.unique(
            [remove_suffixes(x) for x in self.standard.colnames]
        )
        self._colnames["with uncertainties"] = np.unique(
            [
                remove_suffixes(x)
                for x in self.standard.colnames
                if ends_with_any_of(x, uncertainty_suffixes)
            ]
        )
        self._colnames["without uncertainties"] = np.unique(
            [
                x
                for x in self._colnames["everything"]
                if x not in self._colnames["with uncertainties"]
            ]
        )
        # self._colnames["with limits"] = np.unique(
        #    [
        #        remove_suffixes(x)
        #        for x in self.standard.colnames
        #        if ends_with_any_of(x, limit_suffixes)
        #    ]
        # )
        self._colnames["with references"] = np.unique(
            [
                remove_suffixes(x)
                for x in self.standard.colnames
                if ends_with_any_of(x, reference_suffixes)
            ]
        )

    def add_column(
        self,
        name="",
        data=None,
        uncertainty=None,
        uncertainty_lower=None,
        uncertainty_upper=None,
    ):
        """
        Add a column to the existing population.

        This wrapper adds a new column of data to the `.standard` table,
        populates columns for uncertainties if provided, and registers
        a new method for accessing that data.

        Parameters
        ----------
        name : str
            The name of this column; it must be able to be a valid Python variable name.
            If "x", we will add the column `.standard["x"]` and `.x()` as methods.
        data : astropy.units.Quantity
            An array of data to be added into this column. It must have the same
            size at the population, and its entries should be ordered the same
            as the population.
        uncertainty : astropy.units.Quantity
            (optional) Symmetric uncertainties associated with the column.
            If included
        uncertainty_lower : astropy.units.Quantity
            (optional, along with uncertainty_upper) The magnitude of the
            lower uncertainty, for asymmetric uncertainties.
        uncertainty_upper : astropy.units.Quantity
            (optional, along with uncertainty_lower) The magnitude of the
            upper uncertainty, for asymmetric uncertainties.
        """

        # warn if overwriting an existing column
        if name in self.standard.colnames:
            warnings.warn(
                f"'{name}' already exists in `.standard`; you're overwriting something!"
            )

        # put the data into the column
        self.standard[name] = data

        # add uncertainties, if provided
        if uncertainty is not None:
            self.standard[f"{name}_uncertainty"] = uncertainty
        elif (uncertainty_lower is not None) and (uncertainty_upper is not None):
            self.standard["{name}_uncertainty_lower"] = uncertainty_lower
            self.standard["{name}_uncertainty_upper"] = uncertainty_upper

        # register the method for the column
        if hasattr(self, name):
            warnings.warn(
                f"'.{name}()' already exists in `.standard` for this Population; please consider a different name!"
            )
        setattr(self, name, self._create_function_to_access_table_quantity(name))

    def add_calculation(self, name, function):
        """
        Add a new calculation to this population.

        This wrapper adds a new method to calculate a new quantity,
        defining a new function that can be called with access to the
        population's internal data. Uncertainty propagation can also
        be applied to any new calculations registered using
        `.add_calculation`.

        Parameters
        ----------
        name : str
            The name of this calculation; it must be able to be a valid Python variable name.
            If "x", we will add `.x()` as as a new method.
        function : function
            Python function that calculates a quantity. This function
            should look something like:

                def f(self, distribution=False, **kw):
                    '''
                    Brief Name for Quantity (unit)

                    A more detailed description of the quantity,
                    making the docstring for the function more
                    readable for everyone who might use it.
                    '''
                    x = self.yay(distribution=distribution)
                    y = self.wow(distribution=distribution)
                    return x*y

            where the `distribution` keywords enable uncertainty propagation.
        """
        if hasattr(self, name):
            warnings.warn(
                f"Eep! You're overwriting `.{name}()`. That might be OK, but you should be aware!"
            )

        setattr(self.__class__, name, function)

    def print_column_summary(self):
        """
        Print a summary of columns that come directly from the `.standard` table.
        """
        print(f"The following columns are present in the internal `.standard` table:\n")
        for k in self._colnames:
            print(f"{k} =\n{self._colnames[k]}\n")

    def _list_table_indices(self):
        """
        Return a list of keys being used as table indices.

        Core populations should likely have just ['tidyname', 'tidyhostname']
        but internal `.individual_references` populations could have lots,
        one for each quantity with an reference that might be chosen.

        Returns
        -------
        index_keys : list
            The list of keys that are being used as an index.
        """
        return [x.columns[0].name for x in self.standard.indices]

    def _make_sure_index_exists(self, k):
        """
        Add a new key as an index, but don't add it twice.

        Parameters
        ----------
        k : str
            The new key to add.
        """
        if k not in self._list_table_indices():
            self.standard.add_index(k)

    @property
    def _fileprefix(self):
        """
        Define a fileprefix for this population, to be used
        for setting the filename of the standardized population.

        Return
        ------
        """
        return clean(self.label)

    @property
    def _standardized_data_path(self):
        """
        Define the filepath for the standardized table.
        """
        return os.path.join(directories["data"], f"standardized-{self._fileprefix}.txt")

    def save(self, filename=None, overwrite=True):
        """
        Save this population out to a file.

        This saves the standardized data table from this population
        out to a file, along with metadata needed to be loaded
        back into as a drop-in replacement for a live population.

        Parameters
        ----------
        filename : str
            The filepath to which the population should be saved.
        overwrite : bool
            Whether to automatically overwrite existing populations.

        Examples
        --------
        >>> from exoatlas import *
        >>> one = Exoplanets()['GJ1132b']
        >>> one.save('pop.txt')
        >>> the_same_one = Population('pop.txt')
        """

        # be a little fussy about overwriting automatic filenames
        if filename == None:
            filename = f"exoatlas-population-{self._fileprefix}.ecsv"
            overwrite = False
            if os.path.exists(filename):
                warnings.warn(
                    f"{filename} will not be overwritten unless you explicitly provide a filename."
                )

                # save the table as an ascii table for humans to read
        to_save = copy.deepcopy(self.standard)
        to_save.meta["label"] = self.label
        to_save.meta["plotkw"] = self._plotkw

        to_save.write(filename, format="ascii.ecsv", overwrite=overwrite)
        print(
            f"""
        Saved {self} to {filename}.
        It can be reloaded with `x = Population('{filename}')`
        """
        )

    def sort(self, x, reverse=False):
        """
        Sort this population by some key or attribute.

        This sorts the population in place, meaning that the
        Population object from which it is called will be modified.
        Nothing will be returned.

        Parameters
        ----------
        x : str
            The key by which to sort the table.
        reverse : bool
            Whether to reverse the sort order.
                `reverse == False` means low to high
                `reverse == True` means high to low
        """

        # get the values by which to sort the population
        to_sort = getattr(self, x)

        # define the sorting indices
        i = np.argsort(to_sort)
        if reverse:
            i = i[::-1]

        # reorder the standardized table
        self.standard = self.standard[i]

    def __add__(self, other):
        """
        Create a new population by adding two together:

            `bigger = this + other`

        Parameters
        ----------
        other : Population
            The population to be tacked onto this one.

        Returns
        -------
        bigger : Population
            A new population, consisting of all the planets
            in `this` population and some extra ones added
            from `other`.

        """

        # skip any warnings that pop up
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            #  create a new table, joining both together
            table = join(
                self.standard.filled(), other.standard.filled(), join_type="outer"
            )
            # TO-DO, I'm not 100% sure why the tables need to be `filled()` here; shouldn't they already be?

            # create an informative label
            label = f"{self.label} + {other.label}"

        # create and return the new population
        return type(self)(standard=table, label=label)

    def _remove_by_key(self, other, key="tidyname"):
        """
        Create a new population by removing some rows from here:

            `smaller = this - other`

        Parameters
        ----------
        other : Population
            The population of planets to be removed from
            `this` population to create a new `smaller` one.

        Returns
        -------
        smaller : Population
            A subset of `this` population, where some rows
            have been removed.
        """

        # skip any warnings that pop up
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            #  create a new table, joining both together
            table = setdiff(self.standard, other.standard, keys=key)

            # create an informative label
            label = f"{self.label} - {other.label}"

        # create and return the new population
        return type(self)(standard=table, label=label)

    def __sub__(self, other):
        """
        Create a new population by removing some rows from here:

            `smaller = this - other`

        Parameters
        ----------
        other : Population
            The population of planets to be removed from
            `this` population to create a new `smaller` one.

        Returns
        -------
        smaller : Population
            A subset of `this` population, where some rows
            have been removed.
        """
        return self._remove_by_key(other)

    def __getitem__(self, key):
        """
        Create a subpopulation of planets by indexing, slicing, or masking.

        Parameters
        ----------
        key : int, list, array, slice
            An index, slice, or mask to select a subset of the population.

        Returns
        -------
        subset : Population
            A subset of the population, as set by the index, slice, or mask.
        """

        try:
            # if the key is an index/slice/mask, return it
            if self.label is None:
                label = None
            else:
                label = f"Subset of {self.label}"
            subset = type(self)(
                standard=self.standard[key], label=label, **self._plotkw
            )

            # if the key is a column, raise an error
            if type(key) in self.standard.colnames:
                raise IndexError(
                    f"""
                You seem to be trying to access a column from this
                population via `pop[{key}]`. For clarity, all `[]`
                indexing is reserved for selecting subsets of the
                population.

                To access your particular column, please try either
                `pop.{key}` or `pop.standard[{key}]` to return a
                1D array of the entries in that column.
                """
                )
        except KeyError:
            # use a string or a list of strings make a subset by planet name
            # FIXME - maybe we should make this say more when it's making a sneaky choice for us?
            try:
                subset = self.create_subset_by_name(key)
            except KeyError:
                subset = self.create_subset_by_hostname(key)

        # kind of kludgy, to enforce different colors
        subset._plotkw["color"] = None
        subset._plotkw["c"] = None

        return subset

    def create_subset_by_name(self, key):
        """
        Extract a subset of this population,
        based on one or more planet names.

        Parameters
        ----------
        key : strings, list of strings
            The name of a planet ("GJ1132b")
            or a list of planet names.

            (All names will be stripped of
            special characters and converted
            to lower case before indexing.)

        Returns
        -------
        subset : Population
            A new population containing
            some subset of the original.
        """

        # use a (list of) string(s) to index population by name
        if isinstance(key, str):
            # is it just one name?
            tidy = clean(key).lower()
        elif isinstance(key[0], str):
            # is it a list of names?
            tidy = [clean(k).lower() for k in key]

        # pull out rows by planet name
        subset = self.standard.loc["tidyname", tidy]

        # create a useful label for the population
        if isinstance(key, str):
            label = key
        elif isinstance(key[0], str):
            label = "+".join(key)

        # create that new sub-population
        return type(self)(standard=subset, label=label, **self._plotkw)

    def create_subset_by_hostname(self, key):
        """
        Extract a subset of this population,
        based on one or more planet hostnames.

        Parameters
        ----------
        key : strings, list of strings
            The hostname of a planet ("GJ1132")
            or a list of planet hostnames.

            (All names will be stripped of
            special characters and converted
            to lower case before indexing.)

        Returns
        -------
        subset : Population
            A new population containing
            some subset of the original.
        """

        # use a string or a list of strings to index the population by name
        if isinstance(key, str):
            # is it just one name?
            tidy = clean(key).lower()
        elif isinstance(key[0], str):
            # is it a list of names?
            tidy = [clean(k).lower() for k in key]

        # pull out rows by planet name
        subset = self.standard.loc["tidyhostname", tidy]

        # create a useful label for the population
        if isinstance(key, str):
            label = key
        elif isinstance(key[0], str):
            label = "+".join(key)

        # create that new sub-population
        return type(self)(standard=subset, label=label, **self._plotkw)

    def create_subset_by_position(
        self,
        coordinates,
        radius=1 * u.arcmin,
        use_proper_motions=False,
        return_indices=False,
    ):
        """
        Extract a subset of this population,
        by performing a spatial cross-match by
        RA and Dec. This will return all planets
        from this population that fall within
        the specified radius of at least one of
        the specified coordinates.

        Parameters
        ----------
        coordinates : astropy.coordinates.SkyCoord
            The sky coordinate (RA, Dec) or list
            of coordinates we want to search for
            nearby objects.

        radius : astropy.units.Quantity
            The angular radius around each position
            that we should include in each search.

        use_proper_motions : bool
            Should we use available proper motions,
            embedded in the skycoords, to propagate
            positions to a shared epoch before
            cross-matching? Alas, this ability
            is *not yet implemented*. FIXME!

        return_indices : bool
            Should we also return the indices
            of the original coordinates that
            were matched to existing positions?

        Returns
        -------
        subset : Population
            A new population containing a subset
            of the original, including *all* planets
            that fall within the 2D sky search space.

        """

        if use_proper_motions:
            raise NotImplementedError("No cross-matching with proper motions yet :-(")

        # create astropy coordinates for this population
        population_coords = SkyCoord(ra=self.ra(), dec=self.dec())

        # do a spatial cross match on the sky
        #  (idx gives the index into coordinates,
        #   each corresponding to an entry in population_coords)
        idx, d2d, d3d = population_coords.match_to_catalog_sky(coordinates)

        # identify which systems are actually close on the sky
        match = d2d < radius

        # create new populations that are linked by spatial position
        i_match = match.nonzero()[0]
        # matched_coordinates = coordinates[idx[i_match]]
        subset = self.standard[i_match]

        # define a meaningful label
        label = f"Spatial Cross-Match ({len(coordinates)} positions, {radius} radius)"

        # create that new sub-population
        new_population = type(self)(standard=subset, label=label, **self._plotkw)

        # choose what to return
        if return_indices:
            i_from_original_coordinates = idx[i_match]
            return new_population, i_from_original_coordinates
        else:
            return new_population

    def get_uncertainty_lowerupper_from_table(self, key):
        """
        Return two arrays of lower and upper uncertainties, directly from table columns.

        If no uncertainties of any kind are found, a KeyError should be raised.

        Parameters
        ----------
        key : str
            The quantity for which we want lower + upper uncertaintes.

        Returns
        -------
        lower : np.array, u.Quantity
            The magnitude of the lower uncertainties (x_{-lower}^{+upper})
        upper : np.array, u.Quantity
            The magnitude of the upper uncertainties (x_{-lower}^{+upper})
        """

        # this is a bit of a kludge, but when doing error propagation it
        # might be possible for this to get called with either 'stellar_luminosity'
        # or 'stellar_luminosity_from_table'. To avoid triggering a KeyError,
        # we should make sure to strip '_from_table' from the key:
        key = key.replace("_from_table", "")

        # first try for asymmetric table uncertainties
        try:
            lower = _clean_column(self.standard[f"{key}_uncertainty_lower"])
            upper = _clean_column(self.standard[f"{key}_uncertainty_upper"])
            return np.abs(lower), np.abs(upper)
        except KeyError:
            # next try for symmetric table uncertainties
            sym = _clean_column(self.standard[f"{key}_uncertainty"])
            return np.abs(sym), np.abs(sym)

    def get_uncertainty_from_table(self, key, **kw):
        """
        Return an array of symmetric uncertainties on a quantity,
        directly from table columns.

        This (very crudely) averages the lower and upper uncertainties
        to make a symmetric errorbar. For quantities with nearly symmetric
        errors this should be totally fine, but for ones with wildly
        asymmetric uncertainties, use `.get_uncertainty_lowerupper_from_table()`.
        This returns an array, not a function.

        Parameters
        ----------
        key : str
            The quantity for which we want uncertaintes.

        Returns
        -------
        uncertainty : Quantity
            The (symmetrically-averaged) uncertainties, as an array with units.
        """

        sigma_lower, sigma_upper = self.get_uncertainty_lowerupper_from_table(key, **kw)
        return 0.5 * (sigma_lower + sigma_upper)

    def get_values_from_table(self, key, distribution=False):
        """
        Retrieve values directly from the standardized table.

        This wrapper extracts values from the internal `.standard`
        table without performing any calculations or filtering.
        Some quantities might have explicit methods that override
        direct retrieval from the table, but this provides direct
        access to the table values no matter what.

        Parameters
        ----------
        key : str
            The quantity to extract. This must exactly match a
            column in `.standard`; if not, `KeyError` will be raised.
        distribution : bool
            Should it be returned as an astropy distribution,
            for uncertainty propagation?

        Returns
        -------
        values : array, astropy.units.Quantity, astropy.uncertainty.Distribution
            The requested table column.
                If it has units, it will be an astropy Quantity.
                If it has no units, it will be an astropy array.
                If `distribution` is True, it will be an astropy Distribution.
        """

        # extract the column from self.standard for this key
        try:
            raw_column = self.standard[key]
        except KeyError:
            raise KeyError(
                f"The column '{key}' wasn't found in {self}'s internal table."
            )

        # return a distribution or an array
        if distribution:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # set the mode of the distribution to be the table value
                mu = _clean_column(raw_column)
                try:
                    # get the lower + upper uncertainties from the table
                    sigma_lower, sigma_upper = (
                        self.get_uncertainty_lowerupper_from_table(key)
                    )
                    samples = make_skew_samples_from_lowerupper(
                        mu=mu,
                        sigma_lower=sigma_lower,
                        sigma_upper=sigma_upper,
                        N_samples=self._number_of_uncertainty_samples,
                    )
                    return Distribution(samples)
                except KeyError:
                    # if uncertainties fail, give up and return simple array
                    return mu
        else:
            return _clean_column(raw_column)

    def __getattr__(self, key):
        """
        If an attribute/method isn't explicitly defined for a population,
        try to access it as `population.something` with this wrapper.

        Lots of data is stored inside an exoatlas Population; this hidden
        method will be called to try to access those data only if no other
        explicit definition defines that variable first. If we try to access
        an internal Population variable via `pop.x` or `getattr(pop, 'x')`,
        here's what happens, shown approximately in order of precedence:

            1) We look for an explicit definition that has been
            attached to the Population as a method/attribute. This may have been
            defined either with a `def x()` definition inside a module in
            `exoatlas/populations/calculations/` or an attribute/method that
            gets defined anywhere with `pop.x =`, potentially including
            by a user creating a new quantity function and assigning it.
            These definitions should show up with `pop.<tab>` in jupyter.

            2) We look for an implicit method definition that was created
            by `_populate_column_methods()`, where every data column in
            the `.standard` table gets its own function with basic docstring.
            These definitions should also show up with `pop.<tab>` in jupyter.

            3) We use this `__getattr__` function. It will be called only if
            no other definition anywhere overrides it. Practically, this is
            mostly used for retrieving obvious plotting keywords, which
            might be hidden inside the `._plotkw` internal dictionary, or
            to get uncertainties when requesting `.x_uncertainty()` or
            `.x_uncertainty_lowerupper()`.

        Parameters
        ----------
        key : str
            The attribute we're trying to get.

        Returns
        -------
        value : any
            The attribute we're trying to get.
        """

        # do a quick check that something essential isn't missing
        if key in ["label", "_plotkw", "standard"]:
            raise RuntimeError(
                f"""
                Yikes! It looks like `.{key}` isn't defined for this Population. 
                Ideally, this error should never been seen, but if it does, something's 
                gone dreadfully wrong. 
                """
            )

        # try to get a plotkw from this pop, from the plotting defaults, from None
        try:
            assert key in allowed_plotkw
            value = self._plotkw.get(key, None)
            if value is None:
                value = default_plotkw.get(key, None)
            return value
        except (AssertionError, KeyError):
            pass

        # check if we're asking for a quantity explicitly from the table
        if key.endswith("_from_table"):
            quantity_key = key.split("_from_table")[0]

            def f(distribution=False, **kw):
                return self.get_values_from_table(
                    key=quantity_key, distribution=distribution
                )

            f.__docstring__ = f"""
            A function to return the column '.{quantity_key}'. 

            Parameters 
            ----------
            distribution : bool 
                Should this return a Distribution, for uncertainty propagation?

            Returns
            -------
            value : np.array, u.Quantity, astropy.uncertainty.Distribution
                The values, drawn from the table. 
            """
            return f

        # check if we're asking for an uncertainty
        if key.endswith("_uncertainty_lowerupper"):
            quantity_key = key.split("_uncertainty_lowerupper")[0]

            def f(**kw):
                return self.get_uncertainty_lowerupper(key=quantity_key, **kw)

            f.__docstring__ = f"""
            A function to return the two-sided uncertainty on '.{quantity_key}'. 

            Returns
            -------
            lower : np.array, u.Quantity
                The magnitude of the lower uncertainties (x_-lower^+upper)
            upper : np.array, u.Quantity
                The magnitude of the upper uncertainties (x_-lower^+upper)              
            """
            return f

        if key.endswith("_uncertainty"):
            quantity_key = key.split("_uncertainty")[0]

            def f(**kw):
                return self.get_uncertainty(key=quantity_key, **kw)

            f.__docstring__ = f"""
            A function to return the symmetric uncertainty on '.{quantity_key}'. 

            Returns
            -------
            sigma : np.array, u.Quantity
                The magnitude of the uncertainties, assumed to be symmetric.
            """
            return f

        raise AttributeError(
            f"""
            Alas, there seems to be no way to find `.{key}`
            as a table column, attribute, method, or property of {self}.
            """
        )

    def __setattr__(self, key, value):
        """
        Define what happens when we try to set an attribute via `pop.attr = x`.

        If the keyword is a pre-defined "plotting" keyword in `allowed_plotkw`,
        then we should save it in a special `_plotkw` dictionary. Otherwise,
        the attribute should be set as normal.

        Parameters
        ----------
        key : str
            The attribute we're trying to set.
        value : anything
            The value we're trying to give that attribute.
        """

        if key in allowed_plotkw:
            # store plotting keywords in a separate plotting dictionary
            self._plotkw[key] = value
        else:
            # otherwise, store attributes as normal for objects
            self.__dict__[key] = value

    def __repr__(self):
        """
        How should this object appear as a repr/str?
        """
        return f"✨ {self.label} | {len(self)} elements ✨"

    def get(self, key, **kw):
        """
        Return an array quantity for a Population.

        This returns an array, not a function. For example, stellar
        radius might be retrieved either as `.stellar_radius()`
        or as `.get('stellar_radius')`. Quantities that can take
        keyword arguments can pass those keywords as in
        `.stellar_brightness(wavelength=5*u.micron)` or
        `.get('stellar_brightness', wavelength=5*u.micron)`.


        Parameters
        ----------
        key : str
            The name of the quantity we're trying to retrieve.

        kw : dict
            All additional keywords will be passed to the
            method that retrieves the quantities. One common
            (internal) use might be to pass `distribution=True`,
            to be used for uncertainty propagation.

        Returns
        -------
        value : Quantity
            The values, as an array with units.
        """
        return getattr(self, key)(**kw)

    def get_uncertainty_lowerupper(self, key, **kw):
        """
        Return two arrays of lower and upper uncertainties.

        For table column quantities with uncertainties, this should
        return uncertainties direct from `_uncertainty` or
        `_uncertainty_lower` + `_uncertainty_upper` columns in the table.
        This returns an array, not a function.

        For derived quantities, this uncertainty will be estimated from
        central 68% confidence interval of samples from the calculated
        distribution of the quantity.

        Parameters
        ----------
        key : str
            The quantity for which we want lower + upper uncertaintes.
        kw : dict
            All other keywords will be passed to `.get`.

        Returns
        -------
        lower : np.array
            The magnitude of the lower uncertainties (x_{-lower}^{+upper})
        upper : np.array
            The magnitude of the upper uncertainties (x_{-lower}^{+upper})
        """

        # first try for uncertainties direct from table
        try:
            sigma_lower, sigma_upper = self.get_uncertainty_lowerupper_from_table(key)
            return sigma_lower, sigma_upper
        except KeyError:
            mu = self.get(key, **kw)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                f = self.targeted_fractional_uncertainty_precision
                total_number_of_samples = 1 / f**2
                number_of_iterations = int(
                    np.maximum(
                        np.ceil(
                            total_number_of_samples
                            / self._number_of_uncertainty_samples
                        ),
                        1,
                    )
                )

                sigma_lowers, sigma_uppers = [], []
                for i in range(number_of_iterations):
                    d = self.get(key, distribution=True, **kw)
                    # calculate uncertainties from percentiles if possible...
                    if isinstance(d, Distribution):
                        lower, upper = d.pdf_percentiles(
                            100
                            * np.array(
                                [
                                    0.5 - gaussian_central_1sigma / 2,
                                    0.5 + gaussian_central_1sigma / 2,
                                ]
                            )
                        )
                        sigma_lower, sigma_upper = mu - lower, upper - mu
                    # ...otherwise just force the uncertainties to zero
                    else:
                        sigma_lower, sigma_upper = 0 * mu, 0 * mu
                    sigma_lowers.append(sigma_lower)
                    sigma_uppers.append(sigma_upper)
                average_sigma_lower = np.mean(u.Quantity(sigma_lowers), axis=0)
                average_sigma_upper = np.mean(u.Quantity(sigma_uppers), axis=0)

                def replace_negative(x):
                    bad = x < 0
                    x[bad] = np.nan
                    return x

                return replace_negative(average_sigma_lower), replace_negative(
                    average_sigma_upper
                )

    def get_uncertainty(self, key, **kw):
        """
        Return an array of symmetric uncertainties on a quantity.

        This (very crudely) averages the lower and upper uncertainties
        to make a symmetric errorbar. For quantities with nearly symmetric
        errors this should be totally fine, but for ones with wildly
        asymmetric uncertainties, use `.get_uncertainty_lowerupper()`.
        This returns an array, not a function.

        Parameters
        ----------
        key : str
            The quantity for which we want uncertaintes.

        Returns
        -------
        uncertainty : Quantity
            The (symmetrically-averaged) uncertainties, as an array with units.
        """

        sigma_lower, sigma_upper = self.get_uncertainty_lowerupper(key, **kw)
        return 0.5 * (sigma_lower + sigma_upper)

    def get_fractional_uncertainty(self, key, **kw):
        """
        Return an array of estimates of the fractional uncertainty on a quantity.

        This little wrapper calculates the ratio sigma_x/x, to make it
        easier to select subsets on the basis of something like
        "it's been measured to better than 20% precision".

        Parameters
        ----------
        key : str
            The quantity for which we want fractional uncertainty.
        **kw : dict
            Additional keywords will be passed to the

        Returns
        -------
        fractional_uncertainty : Quantity
            The (symmetrically-averaged) uncertainties, as an array with units.
        """
        x = self.get(key, **kw)
        sigma_x = self.get_uncertainty(key, **kw)
        fractional_uncertainty = sigma_x / x
        return fractional_uncertainty

    def _validate_columns(self):
        """
        Make sure this standardized table has all the necessary columns.
        Summarize the amount of good data in each.
        """

        N = len(self.standard)
        for k in basic_columns:
            try:
                n = sum(self.standard[k].mask == False)
            except AttributeError:
                try:
                    n = sum(np.isfinite(self.standard[k]))
                except TypeError:
                    n = sum(np.atleast_1d(self.standard[k] != ""))
            print(f"{k:>25} | {n:4}/{N} rows = {n/N:4.0%} are not empty")

    def _find_index(self, name):
        """
        Return index of a particular planet in the population.

        ??? = maybe this could/should be replaced with some table cleverness?
        """

        return np.array([clean(name) in clean(x) for x in self.name()]).nonzero()[0]

    def update_values(self, planets, **kwargs):
        """
        Update values for one or more planets.

        This modifies the internal `.standard` table
        to update individual values. This is meant to
        be a tool that can be used to provide alternate,
        better, and/or unpublished planet parameters
        to a population.

        Changes made by this function will only last for
        the duration of the Python session in which they
        are run. If you want them to be more permanent,
        you will need to save out the population to a
        custom curated standardized file.

        Parameters
        ----------
        planets : str, list
            Names of one or more planets to update.
        kwargs : dict
            All other keyword arguments will be passed to update
            the values for the planet(s). Quantities should have
            appropriate units.

            If only one planet eis being updated, quantities should
            each be single values. If N planets are being updated,
            quantities should be (N,)-dimensional arrays in the
            same order as the planets.
        """

        # create `tidyname` keys to index by planet name
        if isinstance(planets, str):
            planets_to_index = [clean(planets).lower()]
        else:
            planets_to_index = [clean(p).lower() for p in planets]

        # extract just the subsection of the table relating to these planets
        i = self.standard.loc_indices[planets_to_index]

        # loop over keyword arguments
        for k, v in kwargs.items():

            # make sure it's a valid keyword value to assign
            # assert k in subset.colnames

            if k[-12:] == "_uncertainty":
                # nudge symmetric uncertainties into asymmetric form
                for suffix, sign in zip(["_lower", "_upper"], [-1, 1]):
                    old = self.standard[i][k + suffix] * 1
                    new = sign * np.abs(v)
                    self.standard[i][k + suffix] = new
                    print(f"{planets_to_index} | {k+suffix}: {old} > {new}")
            else:
                # update value in table
                old = self.standard[i][k] * 1
                new = v
                self.standard[i][k] = new
                print(f"{planets_to_index} | {k}: {old} > {new}")

            # warn if uncertainties should have been provided but weren't
            should_have_uncertainty = f"{k}_uncertainty_lower" in self.standard.colnames
            does_have_uncertainty = (f"{k}_uncertainty" in kwargs) or (
                (f"{k}_uncertainty_lower" in kwargs)
                and (f"{k}_uncertainty_upper" in kwargs)
            )
            if should_have_uncertainty and (does_have_uncertainty == False):
                warnings.warn(
                    f"'{k}' should probably have some uncertainties, which you didn't provide."
                )

    def __len__(self):
        """
        How many planets are in this population?
        """
        return len(self.standard)

    def create_table(
        self,
        desired_columns=[
            "name",
            "radius",
            "relative_insolation",
            "stellar_radius",
            "stellar_teff",
            "ra",
            "dec",
            "distance",
        ],
    ):
        """
        Create an astropy table based on this population,
        using a subset of columns, which may include ones
        that have been calculated as Population properties.

        Parameters
        ----------
        desired_columns : list
            The columns you want to include. Anything that
            can be accessed via Population.??? can be provided
            here as a string.

        Returns
        -------
        table : astropy.table.QTable
            A table, with those columns, in the same order
            as the Population itself.
        """
        # FIXME! need to add method support for arguments

        # create a dictionary with the desired columns
        d = {c: getattr(self, c)() for c in desired_columns}

        # turn that into an astropy Table
        t = QTable(d)

        return t

    def create_planning_table(
        self,
        desired_columns=[
            "name",
            "ra",
            "dec",
            "period",
            "transit_midpoint",
            "transit_duration",
            "radius",
            "relative_insolation",
            "stellar_radius",
            "stellar_teff",
            "distance",
        ],
    ):
        """
        Create an astropy table to help plan transit observations.

        Parameters
        ----------
        desired_columns : list
            The columns you want to include. Anything that
            can be accessed via Population.??? can be provided
            here as a string.

        Returns
        -------
        table : astropy.table.QTable
            A table, with those columns, in the same order
            as the Population itself.
        """
        return self.create_table(desired_columns=desired_columns)

    def _choose_calculation(
        self,
        methods=[],
        distribution=False,
        how_to_choose="preference",
        visualize=False,
        **kw,
    ):
        """
        Choose values element-wise from multiple functions.

        This wrapper helps merge together different calculations
        of the same quantity, choosing sources based on preference,
        whether the calculation is finite, and/or uncertainties.

        Parameters
        ----------
        methods : list of strings
            The names of methods to consider for calculating
            the quantity, in order of preference. The order
            matters most if `how_to_choose=='preference'`
            (see below); if `how_to_choose=='precision'`
            then preference doesn't really matter (except
            in cases where two different options have
            identical fractional uncertainties).
        distribution : bool
            If False, return a simple array of values.
            If True, return an astropy.uncertainty.Distribution,
            which can be used for error propagation.
        how_to_choose : str
            A string describing how to pick values from
            among the different possible calculations.
            Options include:
                'preference' = pick the highest preference
                option that produces a finite value
                'precision' = pick the most precise value,
                base on the calculated mean uncertainty
        visualize : bool
            Should we visualize all the options and what gets chosen?
        **kw : dict
            All other keywords will be passed to the functions
            for calculating quantities. All functions must be able
            to accept the same set of keyword.
        """

        #
        errors_indicating_missing_data = (AttributeError, KeyError)

        # construct a list of arrays of values
        value_estimates = []
        for m in methods:
            try:
                this = self.get(m, distribution=distribution, **kw)
                value_estimates.append(this)
            except errors_indicating_missing_data:
                pass

        if how_to_choose == "preference":
            # loop through options, picking the first finite value
            for i, v in enumerate(value_estimates):
                if i == 0:
                    values = v
                isnt_good_yet = np.isfinite(values) == False
                values[isnt_good_yet] = v[isnt_good_yet]
        elif how_to_choose == "precision":
            # calculate (symmetric) fractional uncertainties for all options
            mu_estimates = []
            for m in methods:
                try:
                    this = self.get(m, distribution=False, **kw)
                    mu_estimates.append(this)
                except errors_indicating_missing_data:
                    pass
            uncertainty_estimates = []
            for m in methods:
                try:
                    this = self.get_uncertainty(m, **kw)
                    uncertainty_estimates.append(this)
                except errors_indicating_missing_data:
                    pass

            fractional_uncertainty_estimates = [
                u / v for v, u in zip(mu_estimates, uncertainty_estimates)
            ]
            # loop through options to select the one with smallest uncertainty
            for i, (v, f) in enumerate(
                zip(value_estimates, fractional_uncertainty_estimates)
            ):
                f[np.isnan(f)] = np.inf
                if i == 0:
                    values = v
                    current_fractional_uncertainty = f
                has_smaller_uncertainty = f < current_fractional_uncertainty
                current_fractional_uncertainty[has_smaller_uncertainty] = f[
                    has_smaller_uncertainty
                ]
                values[has_smaller_uncertainty] = v[has_smaller_uncertainty]
        else:
            raise ValueError(
                f"""
            "{how_to_choose}" is not a valid option choosing from among
            {methods}
            Only "preference" or "precision" are currently allowed. 
            """
            )

        if visualize:
            plt.figure(figsize=(8, 3))
            if distribution:
                for m, v in zip(methods, value_estimates):
                    plt.violinplot(
                        dataset=np.array(v.distribution.T), positions=np.arange(len(v))
                    )
                plt.plot(
                    values.pdf_median(),
                    color="black",
                    marker="o",
                    markerfacecolor="none",
                )
            else:
                for m, v in zip(methods, value_estimates):
                    plt.plot(v, alpha=0.5, label=m, marker=".")
                plt.plot(
                    values,
                    color="black",
                    marker="o",
                    markersize=10,
                    markerfacecolor="none",
                    linewidth=0,
                )
            plt.legend(frameon=False)

        return values

    from .calculations.planetary import (
        semimajoraxis_from_period,
        semimajoraxis_from_transit_scaled_semimajoraxis,
        semimajoraxis,
        scaled_semimajoraxis_from_semimajoraxis,
        scaled_semimajoraxis,
        eccentricity,
        argument_of_periastron,
        transit_impact_parameter_from_inclination,
        transit_impact_parameter,
        insolation,
        relative_insolation,
        log_relative_insolation,
        relative_cumulative_xuv_insolation,
        teq,
        planet_luminosity,
        transit_depth_from_radii,
        transit_depth,
        scaled_radius_from_radii,
        scaled_radius,
        transit_duration_from_orbit,
        transit_duration,
        mass_estimated_from_radius,
        radius_estimated_from_mass,
        kludge_mass,
        kludge_radius,
        kludge_stellar_age,
        surface_gravity,
        density,
        escape_velocity,
        orbital_velocity,
        impact_velocity,
        escape_parameter,
        scale_height,
    )

    from .calculations.stellar import (
        stellar_luminosity_from_radius_and_teff,
        stellar_luminosity,
        distance_modulus,
    )

    from .calculations.observability import (
        angular_separation,
        imaging_contrast,
        transmission_signal,
        emission_signal,
        reflection_signal,
        stellar_brightness,
        stellar_brightness_in_telescope_units,
        depth_uncertainty,
        _get_noise_and_unit,
        depth_snr,
        emission_snr,
        reflection_snr,
        transmission_snr,
    )



from ...imports import *
from ...telescopes import *


def angular_separation(self, distribution=False, **kw):
    """
    Maximum Angular Separation (a/D, arcseconds)

    Calculate a representative maximum angular separation
    from the semimajor axis and distance.

    TO-DO:
    Include eccentricity.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    a = self.semimajoraxis(distribution=distribution)
    D = self.distance(distribution=distribution)

    theta = np.arctan(a / D).to(u.arcsec)

    return theta


def imaging_contrast(self, albedo=2 / 3, phase_function=0.25, distribution=False, **kw):
    """
    Reflected Light Imaging Contrast (unitless)

    Calculate an estimated imaging contrast for reflected light.

    (FIXME - check math + explanation on albedo + phase_function!)

    Parameters
    ----------
    albedo : float
        What's the geometric albedo? Default is 2/3, which (I think)
        is reasonably the maximum for Lambertian scattering from
        a face-on planet.
    phase_function : float
        What fraction of reflected light comes toward us, compared
        to the immediate dayside reflection from direct illumination.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    return (
        phase_function
        * albedo
        * 0.25
        * (
            self.kludge_radius(distribution=distribution)
            / self.semimajoraxis(distribution=distribution)
        ).decompose()
        ** 2
    )


def transmission_signal(self, mu=2.3, kludge=False, distribution=False, **kw):
    """
    Transmission Transit Signal (2*H*Rp/Rs**2, unitless)

    Calculate an estimate of the transit depth of
    one atmospheric scale height transiting in front
    of the star.

    TO-DO:
    Should we be fussy about allowing other equilibrium
    temperatures for calculating the scale height?

    Parameters
    ----------
    mu : float
        Mean molecular weight (default 2.3 for something
        like solar composition in chemical equilibrium)
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    H = self.scale_height(mu=mu, kludge=kludge, distribution=distribution)
    Rp = self.radius(distribution=distribution)
    Rs = self.stellar_radius(distribution=distribution)
    depth = (2 * H * Rp / Rs**2).decompose()
    return depth


def reflection_signal(self, albedo=0.1, distribution=False, **kw):
    """
    Reflection Eclipse Signal (unitless)

    Calculate an estimate of the reflected light eclipse depth
    for a particular albedo.

    TO-DO:
    Check reflected light math!!

    Parameters
    ----------
    albedo : float
        How much incoming light is reflected away to space?
        Default of 0.1 might be reasonable for rocky surfaces
        and/or very absorbing hot Jupiters?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    Rp = self.radius(distribution=distribution)
    a = self.semimajoraxis(distribution=distribution)
    return albedo * 0.25 * (Rp / a).decompose() ** 2


def emission_signal(
    self, wavelength=5 * u.micron, albedo=0, f=1 / 4, distribution=False, **kw
):
    """
    Thermal Emission Eclipse Signal (unitless)

    Calculate an estimate of the thermal emission eclipse depth
    at a particular wavelength, for a planet's equilibrium
    temperature calculated for a particular albedo and f.
    Calculations (falsely!) assume Planck spectra for both
    the star and planet. Default parameters assumed 0 albedo
    and that heat is redistributed uniformly over the entire
    sphere of the planet.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    albedo : float
        Fraction of light reflected instead of emitted.
        See .teq() docstring for details + interpretation.
    f : float
        Heat distribution parameter for dayside temperature.
        See .teq() docstring for details and interpretation.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # create thermal emission sources for both star and planet
    import rainbowconnection as rc

    star = rc.Thermal(
        teff=self.stellar_teff(distribution=distribution),
        radius=self.stellar_radius(distribution=distribution),
    )
    planet = rc.Thermal(
        teff=self.teq(albedo=albedo, f=f, distribution=distribution),
        radius=self.radius(distribution=distribution),
    )

    # calculate the depth as the luminosity ratio
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        depths = planet.spectrum(wavelength) / star.spectrum(wavelength)

    return depths


def stellar_brightness(self, wavelength=5 * u.micron, distribution=False, **kw):
    """
    Stellar Brightness (photons/s/m^2/micron)

    Calculate the brightness of the star as observed from the Earth,
    at a particular wavelength assuming Planck thermal emission,
    based on the star's distance, radius, and effective temperature.

    TO-DO:
    Check consistency between Rs/Teff and Ls.
    Consider implementing model spectra grids?!

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # import some tools for easy cartoon spectra
    import rainbowconnection as rc

    # create source with right temperature, size, distance
    Teff = self.stellar_teff(distribution=distribution)
    Rs = self.stellar_radius(distribution=distribution)
    D = self.distance(distribution=distribution)
    star = rc.Thermal(teff=Teff, radius=Rs).at(distance=D)

    # calculate the energy flux
    flux_in_energy = star.spectrum(wavelength)

    # convert to photon flux
    photon_energy = con.h * con.c / wavelength / u.ph
    flux_in_photons = flux_in_energy / photon_energy

    # return photon flux in good units
    return flux_in_photons.to("ph s^-1 m^-2 micron^-1")


def stellar_brightness_in_telescope_units(
    self, telescope_name="JWST", distribution=False, **kw
):
    """
    Stellar Brightness (photons/telescope/hour)

    Calculate the brightness of the star as observed from the Earth,
    converted into units more relevant to a particular telescope.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # what counts as 1 "telescope unit" (e.g. JWST at R=20 at 5 microns for 1 hour)
    telescope_unit = define_telescope_unit_by_name(telescope_name, **kw)

    # what's the photon flux (photons/m**2/s)
    flux_in_photons = self.stellar_brightness(
        wavelength=telescope_unit.wavelength, distribution=distribution
    )

    # quote the brightness as (for example) gigaphotons/JWST at R=20 at 5 microns in 1 hour
    unit = lotsofphotons_unit / telescope_unit
    return flux_in_photons.to(unit)


def depth_uncertainty(
    self,
    telescope_name="JWST",
    per_transit=False,
    dt=1 * u.hour,
    distribution=False,
    **kw
):
    """
    Transit/Eclipse Depth Uncertainty (unitless)

    Calculate the estimated uncertainty expected for transit
    or eclipse depth measurements with a particular telescope.
    The depth uncertainty assumes photon-limited noise based on
    the telescope (with implied effective collecting area),
    the wavelength (which sets the stellar brightness),
    the spectral resolution (which sets the bandpass width),
    and the time being observed.

    By default, depth uncertainties will be calculated per constant time;
    optionally, depth uncertainty can be calculated per transit,
    where longer transits will have relatively smaller uncertainty.


    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # what counts as 1 "telescope unit" (e.g. JWST at R=20 at 5 microns for 1 hour)
    telescope_unit = define_telescope_unit_by_name(telescope_name, dt=dt, **kw)

    # what's the photon flux (photons/m**2/s)
    flux_in_photons = self.stellar_brightness(
        wavelength=telescope_unit.wavelength, distribution=distribution
    )

    # what's the total collecting power?
    if per_transit:
        ratio_of_collecting_time = self.transit_duration(distribution=distribution) / dt
    else:
        ratio_of_collecting_time = 1
    collecting_power = 1 * telescope_unit * ratio_of_collecting_time

    # what's the total number of photons collected during transit
    N = (flux_in_photons * collecting_power).to(u.ph).value

    # what's the flux uncertainty on the time scale of one transit?
    sigma = 1 / np.sqrt(N)

    # inflate by a factor of sqrt(2) for equal out-of-transit
    oot = np.sqrt(2)
    sigma_depth = sigma * oot

    return u.Quantity(sigma_depth)


def _get_noise_and_unit(
    self, telescope_name="JWST", per_transit=False, distribution=False, **kw
):
    """
    Tiny helper to get the noise and the telescope_unit
    for a telescope observation of a planet.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # figure out the noise
    noise = self.depth_uncertainty(
        telescope_name=telescope_name,
        per_transit=per_transit,
        distribution=distribution,
        **kw
    )

    # create a telescope unit (mostly to get a default wavelength)
    telescope_unit = define_telescope_unit_by_name(telescope_name, **kw)

    return noise, telescope_unit


def depth_snr(self, telescope_name="JWST", distribution=False, **kw):
    """
    Transit Depth S/N

    Calculate the S/N for detecting the transit
    with a particular telescope.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    noise, telescope_unit = self._get_noise_and_unit(
        telescope_name=telescope_name, distribution=distribution, **kw
    )
    signal = self.transit_depth(distribution=distribution)
    return signal / noise


def emission_snr(
    self, telescope_name="JWST", albedo=0, f=1 / 4, distribution=False, **kw
):
    """
    Thermal Emission Eclipse Depth S/N

    Calculate the S/N for detecting the
    planet's thermal emission eclipse
    with a particular telescope.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    albedo : float
        Fraction of light reflected instead of emitted.
        See .teq() docstring for details + interpretation.
    f : float
        Heat distribution parameter for dayside temperature.
        See .teq() docstring for details and interpretation.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    noise, telescope_unit = self._get_noise_and_unit(
        telescope_name=telescope_name, distribution=distribution, **kw
    )
    kw["wavelength"] = telescope_unit.wavelength
    signal = self.emission_signal(albedo=albedo, f=f, distribution=distribution, **kw)
    return signal / noise


def reflection_snr(self, telescope_name="JWST", albedo=0.1, distribution=False, **kw):
    """
    Reflected Light Eclipse Depth S/N

    Calculate the S/N for detecting the
    planet's reflected light eclipse
    with a particular telescope.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    albedo : float
        Fraction of light reflected instead of emitted.
        See .teq() docstring for details + interpretation.
    f : float
        Heat distribution parameter for dayside temperature.
        See .teq() docstring for details and interpretation.
    albedo : float
        How much incoming light is reflected away to space?
        Default of 0.1 might be reasonable for rocky surfaces
        and/or very absorbing hot Jupiters?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    noise, telescope_unit = self._get_noise_and_unit(
        telescope_name=telescope_name, distribution=distribution, **kw
    )
    signal = self.reflection_signal(albedo=albedo, distribution=distribution)
    return signal / noise


def transmission_snr(
    self, telescope_name="JWST", mu=2.3, kludge=False, distribution=False, **kw
):
    """
    Transmission Spectroscopy Transit Depth S/N

    Calculate the S/N for detecting the planet's
    transmission spectrum change in transit depths
    with a particular telescope.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.
    per_transit : bool
        If True, calculate the depth uncertainty for one transit,
        based on the calculated transit duration for the planet.
        If False, calculate the depth uncertainty for a certain amount
        of in-transit time. You likely want to specify `dt` as a
        keyword argument to set that amount of in-transit time.
        In either case, an out-of-transit baseline equal to the
        total in-transit time will be assumed. This means the actual
        time cost will be twice the transit duration or `dt` chosen,
        and the depth uncertainty will be a factor sqrt(2) larger
        than the pure photon noise binned to the relevant timescale.
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.
    R : float
        The spectral resolution at which the
        telescope will bin wavelengths.
    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes. If `per_transit=True`,
        this will be ignored. Otherwise, it will set the total amount
        of in-transit time observed, assuming that an equal amount of
        time will *also* be observed out of transit.
    albedo : float
        Fraction of light reflected instead of emitted.
        See .teq() docstring for details + interpretation.
    f : float
        Heat distribution parameter for dayside temperature.
        See .teq() docstring for details and interpretation.
    mu : float
        Mean molecular weight (default 2.3 for something
        like solar composition in chemical equilibrium)
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    noise, telescope_unit = self._get_noise_and_unit(
        telescope_name=telescope_name, distribution=distribution, **kw
    )
    signal = self.transmission_signal(mu=mu, kludge=kludge, distribution=distribution)
    return signal / noise



"""define a Transit class, to represent an individual planetary transit"""

from ..imports import *


class Transit:
    """a Transit object is one transit of one (associated) planet"""

    def __init__(
        self, planet, i, phasefrommidtransit=0.0, observatory=None, buffer=1.5
    ):
        """
        initialize:
            planet = a Planet object (which contains a period and epoch)
            i = the epoch of this transit
            phasefrommidtransit = set this to 0.5 for eclipses (for circular)
            plan = the Plan object in which this is embedded
        """

        # store lots of links to other objects
        self.planet = planet
        self.coord = SkyCoord(self.planet.ra, self.planet.dec)
        self.observatory = observatory

        # set the midtransit time of this particular transit
        self.midtransit = (
            planet.period * (i + phasefrommidtransit) + planet.transit_midpoint
        )[0]
        self.i = i

        # how much padding should be on the other side of the transit
        self.buffer = buffer

    @property
    def duration(self):
        """the duration of the transit"""

        # this should be in days
        d = self.planet.transit_duration[0]

        if np.isfinite(d):
            return d
        else:
            return 5.0 * u.hour  # KLUDGE!

    @property
    def ingress(self):
        """when is ingress?"""
        return self.midtransit - self.duration * 0.5

    @property
    def egress(self):
        """when is egress?"""
        return self.midtransit + self.duration * 0.5

    @property
    def pretransit(self):
        """when should observations start?"""
        return self.ingress - self.duration * self.buffer

    @property
    def posttransit(self):
        """when should observations end?"""

        return self.egress + self.duration * self.buffer

    @property
    def withbuffer(self):
        """list of plot_date times from start to finish of observation"""
        return [self.pretransit, self.posttransit]

    @property
    def justtransit(self):
        """list of plot_date times from start to finish of transit"""
        return [self.ingress, self.egress]

    def plot(
        self, alpha=0.6, linewidth=5, marker=None, color="black", **kwargs
    ):  # y=0,
        """plot this transit within a night"""

        kwargs["marker"] = marker
        kwargs["linewidth"] = linewidth
        kwargs["alpha"] = alpha
        # kwargs['color'] = self.planet.color

        # plot the transits
        # vertical = [y,y]

        # figure out the airmass for this transit
        times = self.midtransit + np.linspace(-1, 1, 100) * self.duration * self.buffer
        astropy_times = Time(times, format="jd")
        altaz = self.observatory.altaz(self.coord, astropy_times)
        airmass = altaz.secz.value
        ok = altaz.alt > 0

        # why is this here?
        try:
            assert self.duration < 0.5 * u.day
        except AssertionError:
            return

        # pull out the intransit subset
        intransit = (times >= self.ingress) * (times <= self.egress)

        # plot the full airmass curve
        plt.plot(astropy_times.plot_date[ok], airmass[ok], color=color, **kwargs)

        # plot the airmass curve of just the transit
        plt.plot(
            astropy_times.plot_date[ok * intransit],
            airmass[ok * intransit],
            color=color,
            **kwargs,
        )

        # add a label to this transit
        y = np.min(airmass[ok * intransit]) - 0.05
        description = f"{self.planet.name[0]}\n"
        description += (
            rf"{self.planet.transit_depth[0]:.2%},{self.planet.radius[0].value:.1f}$R_\oplus$,{self.planet.stellar_radius[0].value:.1f}$R_*$"
            + "\n"
        )
        description += rf"{self.planet.relative_insolation[0]:.0f}$S_\oplus$,{self.planet.distance[0]:.0f}"
        plt.text(
            Time(self.midtransit, format="jd").plot_date,
            y,
            description,
            color=color,
            ha="center",
            va="bottom",
            weight="bold",
            alpha=kwargs["alpha"],
            fontsize=5,
        )

    def airmass(self, time):
        """airmass string, for a particular time"""
        astropy_time = Time(time, format="jd")
        return "{0:.1f}".format(
            self.observatory.altaz(self.coord, astropy_time).secz.value[0]
        )

    def sunalt(self, time):
        """sunaltitude, for a particular time"""
        astropy_time = Time(time, format="jd")

        return "{0:+3.0f}".format(self.observatory.sun(astropy_time).alt.deg)

    def simpletime(self, time):
        astropy_time = Time(time, format="jd")
        return "{0:.5f} = {1} UT (A={2},S={3})".format(
            astropy_time.jd,
            astropy_time.iso[0:16],
            self.airmass(astropy_time),
            self.sunalt(astropy_time),
        )

    def details(self):
        s = f"{self.planet.name[0]:<20} \n"  # [not accounting for heliocentric]
        s += " {0:3.1f}x duration before = {1}\n".format(
            self.buffer, self.simpletime(self.pretransit)
        )
        s += "              ingress = {0}\n".format(self.simpletime(self.ingress))
        s += "          mid-transit = {0}\n".format(self.simpletime(self.midtransit))
        s += "               egress = {0}\n".format(self.simpletime(self.egress))
        s += "  {0:.1f}x duration after = {1}\n".format(
            self.buffer, self.simpletime(self.posttransit)
        )

        print(s)
        return s
        # print('mid-transit = {0:.5f}  [{1} to {2} UT, secz = {3:.2f} to {4:.2f}, sun = {5:+3.0f} to {6:+3.0f}]'.format(

        #                    self.observatory.sun(t.pretransit).alt.deg,
        #                    self.observatory.sun(t.posttransit).alt.deg))



"""
Tools to summarize a particular population.
"""

from ..imports import *
from .column_descriptions import *

__all__ = ["plot_histograms", "summarize_planet"]


def split_cols(pop):
    # find the columns that aren't strings
    quant, qual = [], []
    for x in pop.standard.colnames:
        try:
            pop.standard[x][0] + "a"
            qual.append(x)
        except TypeError:
            quant.append(x)
    return quant, qual


def plot_histograms(pop):
    """
    Make histograms of all the necessary columns in the population.
    These can help see what's missing and what's there.
    """

    # pull out the quantitative columns
    quant_cols, string_cols = split_cols(pop)

    cols = []
    for x in string_cols:
        if len(np.unique(pop.standard[x])) < 50:
            cols.append(x)
    cols += quant_cols

    # create a grid of histograms
    ncols = 3
    nrows = np.ceil(len(cols) / ncols).astype(np.int)
    scale = 2
    fi, ax = plt.subplots(
        nrows,
        ncols,
        figsize=(ncols * 2 * scale, nrows * scale),
        constrained_layout=True,
    )

    for i, x in enumerate(cols):
        if x in quant_cols:
            good = np.isfinite(pop.standard[x])
        else:
            good = np.ones(len(pop.standard)).astype(bool)
        bad = good == False
        badfraction = sum(bad) / len(bad)

        plt.sca(ax.flatten()[i])
        plt.hist(pop.standard[x][good], color="black")
        plt.axvspan(*plt.xlim(), 0, badfraction, color="red", alpha=0.5, zorder=-1)
        plt.xlabel(x)
        plt.title(f"{x} lacks {sum(bad)}/{len(bad)} ({badfraction:.0%})")


def summarize_planet(planets):
    """
    Quick tool to print out everything we need about a planet.
    """
    for p in planets:
        for k in all_columns:
            print(f"{k:>20} = {getattr(p, k)()}")



from .setup_tests import *

from exoatlas.models import *
from exoatlas.imports import *


def test_seager():
    """
    Test the Seager mass-radius relations.
    """
    plt.cla()
    plot_both_seager()


def test_zeng():
    plt.cla()
    plot_three_zeng()


def test_chen():
    plt.cla()
    plot_chen("mass")
    plot_chen("radius", alpha=0.5, linewidth=4)


def test_kopparapu():
    # test a default HZ
    T = np.linspace(2600, 7000)
    f = make_hz()
    f(T)

    # test that it breaks when it should
    with pytest.raises(Exception):
        make_hz("something-that-does-not-exist")


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



"""
Define an observing Block. This might be a particular
semester/trimester/quarter linked to an observing cycle,
or maybe it's just that you want to know what's up for
the next week or so.
"""

from ..imports import *


class Block:
    """
    The Block object, for generating a sequence of nights.

    :param start:
        The start of the block ()

    """

    def __init__(self, start=None, finish=None, plan=None, name="thisweek"):
        """
        Parameters
        ----------
        start : astropy.time.Time
            The start of the time range to consider.
            (defaults to now)
        finish : astropy.time.Time
            The end of the time range to consider.
            (defaults to start + 7 days)
        plan : Plan
            The plan to which this block is connected.
        name : str
            A descriptive name for this time range.
        """

        # store the connection to the plan
        self.plan = plan

        try:
            assert type(start) == Time
            assert start is not None
        except AssertionError:
            start = Time.now()

        try:
            assert type(finish) == Time
            assert finish is not None
        except AssertionError:
            finish = start + 7 * u.day

        # set the start and finish times (floored down to the start of the day), as JD
        zone = self.plan.observatory.standardzone.to("day").value
        self.start = np.floor(start.jd) + zone
        self.finish = np.ceil(finish.jd) + zone

        # set the noons (before the night = 0.0) and midnights to consider (0.5)
        self.noons = (np.arange(self.start, self.finish)) * u.day
        self.midnights = self.noons + 0.5 * u.day

        # maybe these could be gotten rid of?
        resolution = 5.0 / 60.0 / 24.0
        self.times = np.arange(self.start - 1, self.finish + 2.0, resolution)

        # n = (self.finish - self.start + 2*u.day)/resolution
        # self.start + np.arange(n)*resolution - 1*u.day

        self.name = name



"""
Define some commonly useful subsets of all transiting exoplanets,
giving them unique names and colors to simplify plotting
"""

from ...imports import *
from .transiting_exoplanets import *

__all__ = [
    "TransitingExoplanetsSubset",
    "Kepler",
    "NonKepler",
    "TESS",
    "NonTESS",
    "Space",
    "Ground",
    "GoodMass",
    "BadMass",
]


class TransitingExoplanetsSubset(TransitingExoplanets):
    def __init__(self, label="Subset", **kw):
        TransitingExoplanets.__init__(self, **kw)

        # set the label
        self.label = label

        # trim to just the data we want
        self.standard = self.standard[self.to_include()]

        self._plotkw["color"] = None
        self._plotkw["c"] = None

    def to_include(self):
        raise NotImplementedError(
            "Please define `.to_include()` for this ExoplanetSubset!"
        )


class Kepler(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="Kepler", color="royalblue", zorder=0, **kw
        )

    def to_include(self):
        foundbykepler = (self.discovery_facility() == "Kepler") | (
            self.discovery_facility() == "K2"
        )
        return foundbykepler


class NonKepler(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="Non-Kepler", color="black", zorder=0, **kw
        )

    def to_include(self):
        foundbykepler = (self.discovery_facility() == "Kepler") | (
            self.discovery_facility() == "K2"
        )
        return foundbykepler == False


class TESS(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="TESS", color="orangered", zorder=0, **kw
        )

    def to_include(self):
        foundbytess = (
            self.discovery_facility() == "Transiting Exoplanet Survey Satellite (TESS)"
        )
        return foundbytess == True


class NonTESS(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="NonTESS", color="black", zorder=0, **kw
        )

    def to_include(self):
        foundbytess = (
            self.discovery_facility() == "Transiting Exoplanet Survey Satellite (TESS)"
        )
        return foundbytess == False


space_telescopes = [
    "Transiting Exoplanet Survey Satellite (TESS)",
    "K2",
    "Kepler",
    "CoRoT",
    "Hubble Space Telescope",
]


class Space(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="Space-based", color="orchid", zorder=0, **kw
        )

    def to_include(self):
        foundfromspace = np.zeros(len(self)).astype(bool)
        for x in space_telescopes:
            foundfromspace = foundfromspace | (self.discovery_facility() == x)
        return foundfromspace


class Ground(TransitingExoplanetsSubset):
    def __init__(self, **kw):
        TransitingExoplanetsSubset.__init__(
            self, label="Ground-based", color="black", zorder=0, **kw
        )

    def to_include(self):
        foundfromspace = np.zeros(len(self)).astype(bool)
        for x in space_telescopes:
            foundfromspace = foundfromspace | (self.discovery_facility() == x)
        return foundfromspace == False


sigma = 3


def mass_is_good(pop):
    with np.errstate(invalid="ignore"):
        # the uncertainty must be greater than 0
        exists = pop.get_uncertainty("mass") > 0

        # the uncertainty must be less than a maximum
        fractional = pop.get_uncertainty("mass") / pop.get("mass")
        small = fractional < pop.maximum_uncertainty

        return small & exists


class GoodMass(TransitingExoplanetsSubset):
    def __init__(self, sigma=sigma, **kw):
        self.maximum_uncertainty = 1 / sigma
        TransitingExoplanetsSubset.__init__(self, label="Good Mass", **kw)

    def to_include(self):
        return mass_is_good(self)


class BadMass(TransitingExoplanetsSubset):
    def __init__(self, sigma=sigma, **kw):
        self.maximum_uncertainty = 1 / sigma
        TransitingExoplanetsSubset.__init__(
            self, label="Bad Mass", color="lightblue", **kw
        )

    def to_include(self):
        return mass_is_good(self) == False



from .Map import *
from .BubbleMap import BubbleMap
from ..ink_errorbar import *

__all__ = ["ErrorMap"]


class ErrorMap(BubbleMap):
    """
    Error is a general wrapper for making scatter plots
    where planets are represented with 2D error bars, with
    their intensity scaled to some overall visual weight.
    """

    def intensity(self, invisible_fraction=0.75, x_power=2, y_power=2):
        """
        What visual intensity should each datapoint have?

        By default, this will be set by the product of the
        fractional uncertainties on both x and y.

        Parameters
        ----------
        invisible_fraction : float
            The aproximate 2D fractional uncertainty at which
            a point will fade to invisible. For example, if
            set to 0.75, then points will disappear when
              (sigma_x/x)**2 + (sigma_y/y)**2 > 0.75**2

        x_power : float
            The power to which x is raised in the quantity
            used to define the visual weight. For example,
            if we want visual weight to scale with the
            fractional uncertainty on density (mass/radius**3)
            for x showing mass and y showing radius, we'd
            choose x_power=1 and y_power=3.

        y_power : float
            (see `x_power` above)

        Returns
        -------
        intensity : np.array
            The intensity of the points for the current
            population, as a numeric value between 0 and 1.
            By default,
        """

        dlnx = self.x_uncertainty / self.x * np.abs(x_power)
        dlny = self.y_uncertainty / self.y * np.abs(y_power)

        # things with bigger errors should have lower weight
        weight = 1 - np.sqrt(dlnx**2 + dlny**2) / invisible_fraction

        # clip the weights above 1 (shouldn't exist?) or below zero
        # clipped = np.minimum(np.maximum(weight, 0), 1)

        # return the visual weight
        return remove_unit(weight)

    def plot(self, pop, ax=None, annotate_kw={}, **kw):
        """
        Add the errorbars for a particular population to this map.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        ax :
            Into what ax should we place this plot?
            If None, create a new plotting axes.
        annotate_kw : dict
            Keywords for labeling the planet names.
        **kw : dict
            Any extra keywords will be passed on to `errorbar`
        """

        # focus attention on that population
        self.point_at(pop)

        # do a bubble instead, if requested (for drawing individual systems?)
        if getattr(self.pop, "bubble_anyway", False):
            # FIXME - maybe change to just Map?
            BubbleMap.plot(self, pop=pop, ax=ax, annotate_kw=annotate_kw, **kw)
            return

        # make sure we're plotting into the appropriate axes
        self.setup_axes(ax=ax)

        # define the data we're trying to plot
        x = remove_unit(self.x)
        y = remove_unit(self.y)

        # set the base color to use throughout
        default_color = plt.scatter([], []).get_facecolor()[0]
        color = self.pop.color or default_color
        marker = self.pop.marker or "o"

        # if the entire population is exact (e.g., Solar System),
        # then don't include any errors when plotting
        if self.pop.exact:
            # define plotting keywords without errorbars
            scatterkw = self.kw(color=color, marker=marker, **kw)

            self.scattered[self.pop_key] = plt.scatter(x, y, **scatterkw)
            # FIXME, 5/25/20: I think BubbleMap is doing
            # something a little more clever with being able
            # to manage face and edge colors separately.
            # Perhaps we should set things up so that we might
            # inherit some of this skills here in ErrorMap
        else:
            # define the error bars to be plotting
            xl, xu = self.x_uncertainty_lowerupper
            x_unc = remove_unit(np.vstack([xl, xu]))

            yl, yu = self.y_uncertainty_lowerupper
            y_unc = remove_unit(np.vstack([yl, yu]))

            width = 1
            kw = dict(  # marker='o',
                linewidth=0,
                elinewidth=width,
                alpha=1.0,
                # capthick=width,
                # capsize=2,
                # markersize=3)
                # color=self.pop.color,
                # markeredgecolor=self.pop.color,
            )

            # define an Nx4 array of RGBA colors for the N points
            weights = self.intensity()

            # remove everything with infinite errorbars
            ok = (
                np.isfinite(xl)
                & np.isfinite(xu)
                & np.isfinite(x)
                & np.isfinite(yl)
                & np.isfinite(yu)
                & np.isfinite(y)
                & np.isfinite(weights)
            )

            n_nouncertainty = sum(ok == False)
            # print(
            #    f"skipping {n_nouncertainty} planets that are missing data or uncertainties"
            # )

            # kludge to remove those that cross zero
            with np.errstate(invalid="ignore"):
                ok *= (self.x - xl) > 0
                ok *= (self.y - yl) > 0
            # FIXME, 5/25/2020: This kludge always useful on
            # logarithmic axes (which are the only that have
            # been defined so far), but at some point we
            # might want to use linear axes too, where we might
            # not want to throw out values that might go
            # negative.

            n_consistentwithzero = sum(ok == False) - n_nouncertainty
            # print(
            #    f"skipping {n_consistentwithzero} planets that are consistent with zero"
            # )

            if (len(x) > 1) & (self.pop._plotkw.get("ink", True)):
                # print("plotting inked errorbars, this may take a while")
                # FIXME, 5/25/2020: We should make the
                # "invisible" color be something more flexible
                # than white, in case we're plotting on a dark
                # background. Remember, things look messy if
                # we use alpha to do the visual weighting for
                # these errorbars, because it introduces many
                # more intersecting lines.
                self.scattered[self.pop_key] = ink_errorbar(
                    x[ok],
                    y[ok],
                    yerr=y_unc[:, ok],
                    xerr=x_unc[:, ok],
                    c=weights[ok],
                    cmap=one2another(
                        bottom="white", top=color, alphabottom=1.0, alphatop=1.0
                    ),
                    **kw,
                )
            else:
                self.scattered[self.pop_key] = self.ax.errorbar(
                    x[ok],
                    y[ok],
                    yerr=y_unc[:, ok],
                    xerr=x_unc[:, ok],
                    color=self.pop.color,
                    **kw,
                )

        # set the scales, limits, labels
        self.refine_axes()

        # add planet or hostname labels
        self.add_system_annotations(annotate_kw=annotate_kw)



from ..imports import *
from pytz import timezone
from astropy.coordinates import EarthLocation, get_sun, AltAz
import matplotlib.dates as mdates

# create an empty dictionary
observatories = {}

# define a few observatories
#  parameters copied from skycalc.c source
observatories["CTIO"] = dict(
    name="Cerro Tololo",
    timezone="Chilean",
    standardzone=4.0 * u.hour,
    usedaylightsaving=-1,
    longitude=4.721 * u.hourangle,
    latitude=-30.165 * u.deg,
    elevsea=2215.0 * u.m,
    elev=2215.0 * u.m,  # /* for ocean horizon, not Andes! */
)

observatories["LCO"] = dict(
    name="Las Campanas Observatory",
    timezone="Chilean",
    standardzone=4.0 * u.hour,
    usedaylightsaving=-1,
    longitude=4.71333 * u.hourangle,
    latitude=-29.00833 * u.deg,
    elevsea=2282.0 * u.m,
    elev=2282.0 * u.m,  # /* for ocean horizon, not Andes! */
)

observatories["FLWO"] = dict(
    name="Mt. Hopkins, AZ",
    timezone="Mountain",
    standardzone=7.0 * u.hour,
    usedaylightsaving=0,
    longitude=7.39233 * u.hourangle,
    latitude=31.6883 * u.deg,
    elevsea=2608.0 * u.m,
    elev=500.0 * u.m,  # /* approximate elevation above horizon mtns */
)

observatories["APO"] = dict(
    name="Apache Point Observatory",
    timezone="Mountain",
    standardzone=7.0 * u.hour,
    usedaylightsaving=0,
    longitude=105 * u.deg,
    latitude=32 * u.deg,
    elevsea=2798.0 * u.m,
    elev=500.0 * u.m,  # /* approximate elevation above horizon mtns */
)


observatories["SBO"] = dict(
    name="Sommers-Bausch Observatory",
    timezone="Mountain",
    standardzone=7.0 * u.hour,
    usedaylightsaving=0,
    longitude=105.2635 * u.deg,
    latitude=40.00372 * u.deg,
    elevsea=1653.0 * u.m,
    elev=500.0 * u.m,  # /* approximate elevation above horizon mtns */
)


class Observatory:
    """
    The Observatory object, for keeping track of timing calculating alt-az.

    :param name:
        String listing short name (LCO, CTIO, FLWO) of obsrvatory.

    """

    def __repr__(self):
        return f"<{self.name}>"

    def __init__(self, abbreviation=None, plan=None):
        """
        Initialize an observatory.
        """

        # make sure an abbreviation is defined
        # if abbreviation is None:
        #    print("Pick an observatory from following options:")
        #    print(str(observatories.keys()))
        #    abbreviation = self.input().strip()

        # populate the attributes of this observatory
        o = observatories[abbreviation]
        for k in o.keys():
            self.__dict__[k] = o[k]

        # set this observatory's location on Earth
        self.location = EarthLocation(
            lat=self.latitude, lon=-self.longitude, height=self.elevsea
        )

        # connect this observatory to a plan
        self.plan = plan

    def sun(self, times):
        """
        Get the alt/az of the Sun (at some times).
        """

        sunCoords = get_sun(times)
        altAzFrame = AltAz(obstime=times, location=self.location)
        return sunCoords.transform_to(altAzFrame)

    def altaz(self, coord, times):
        """
        Get the altaz of the star (at some times).
        """

        frame = AltAz(obstime=times, location=self.location)
        return coord.transform_to(frame)

    def plotSun(self, jd, ax=None, threshold=-12.0, color="black"):
        """
        Plot the alitude of the Sun, for some given times.
        """

        print("plotting the sun")

        times = Time(jd, format="jd")
        sunAltAz = self.sun(times)

        dt = times[1:] - times[:-1]
        nudged = times[:-1] + dt / 2.0
        a = sunAltAz.alt.deg - threshold
        sunrises = ((a[1:] > 0) * (a[:-1] < 0)).nonzero()
        sunsets = ((a[1:] < 0) * (a[:-1] > 0)).nonzero()
        starts = nudged[sunrises]
        finishes = nudged[sunsets]
        maxlength = min(len(starts), len(finishes))
        if starts[0] > finishes[0]:
            starts, finishes = starts[: maxlength - 1], finishes[1:maxlength]
        else:
            starts, finishes = starts[:maxlength], finishes[:maxlength]

        assert (finishes > starts).all()
        for i in range(len(starts)):
            plt.axvspan(
                starts[i].plot_date,
                finishes[i].plot_date,
                color=color,
                alpha=0.8,
                zorder=10000,
            )

    def plot_empty_night(self, midnight, threshold=-12.0, ax=None):
        astropy_midnight = Time(midnight, format="jd")

        plt.figure(figsize=(8, 3), dpi=400)
        times = Time(
            np.arange(midnight.value - 1.0, midnight.value + 1.0, 5.0 / 60 / 24),
            format="jd",
        )

        sunAltAz = self.sun(times)

        dt = times[1:] - times[:-1]
        nudged = times[:-1] + dt / 2.0
        a = sunAltAz.alt.deg - threshold
        sunrises = ((a[1:] > 0) * (a[:-1] < 0)).nonzero()
        sunsets = ((a[1:] < 0) * (a[:-1] > 0)).nonzero()
        starts = nudged[sunrises]
        finishes = nudged[sunsets]
        maxlength = min(len(starts), len(finishes))
        if starts[0] > finishes[0]:
            starts, finishes = starts[: maxlength - 1], finishes[1:maxlength]
        else:
            starts, finishes = starts[:maxlength], finishes[:maxlength]

        assert (finishes > starts).all()
        for i in range(len(starts)):
            plt.axvspan(
                starts[i].plot_date,
                finishes[i].plot_date,
                color="black",
                alpha=0.8,
                zorder=10000,
            )

        # set the formatting on the time label
        fmt = plt.matplotlib.dates.DateFormatter("UT %m-%d|%H")
        hours = mdates.HourLocator(interval=1)  #
        ax = plt.gca()
        ax.xaxis.set_major_formatter(fmt)
        ax.xaxis.set_major_locator(hours)
        ax.tick_params(axis="x", labelsize=6.5)
        plt.xlim(astropy_midnight.plot_date - 0.5, astropy_midnight.plot_date + 0.5)
        plt.ylabel("Airmass")

        f = plt.gcf()
        f.autofmt_xdate(rotation=45)

    def plotAirmass(self, coord, times, **kwargs):
        """
        Plot the airmass of the object (at some times).
        """

        print("plotting airmass")
        # calculate altaz
        altaz = self.altaz(coord, times)
        airmass = altaz.secz.value
        ok = altaz.alt > 0

        self.plan.ax["airmass"].plot(times.plot_date[ok], airmass[ok], **kwargs)
        self.plan.ax["airmass"].set_ylim(3, 1)



from ...imports import *


'''def stellar_luminosity_from_table(self, distribution=False, **kw):
    """
    Stellar Luminosity (L*, Lsun)

    Calculate the stellar luminosity from the quoted table value,
    which for the NASA Exoplanet Archive is stored as log10(L/Lsun).

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    logL = self.get("stellar_logluminosity", distribution=distribution)
    L = 10**logL * u.Lsun
    return L'''


def stellar_luminosity_from_radius_and_teff(self, distribution=False, **kw):
    """
    Stellar Luminosity (L*, Lsun)

    Calculate "L*" from the radius and effective temperature.
    This might be used if no stellar luminosity is provided
    in the standardized table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    T = self.get("stellar_teff", distribution=distribution)
    R = self.get("stellar_radius", distribution=distribution)
    sigma = con.sigma_sb
    L = (4 * np.pi * R**2 * sigma * T**4).to(u.Lsun)
    return L


def stellar_luminosity(self, distribution=False, **kw):
    """
    Stellar Luminosity (L*, Lsun)

    Retrieve "L*" first from the standardized table,
    then from radius + effective temperature.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    L = self._choose_calculation(
        methods=[
            "stellar_luminosity_from_table",
            "stellar_luminosity_from_radius_and_teff",
        ],
        distribution=distribution,
        **kw,
    )
    return L


def distance_modulus(self, distribution=False, **kw):
    """
    Distance Modulus ($\mu$, magnitudes)

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    mu = 5 * np.log10(self.distance(distribution=distribution) / (10 * u.pc)) * u.mag
    return mu



from ...imports import *
from ...models import plot_both_seager, make_hz
from ..plottables.preset_plottables import *
from .Map import *
from .BubbleMap import *
from .ErrorMap import *


class Flux_x_Radius(BubbleMap):
    xaxis = Flux
    yaxis = Radius

    def add_teqaxis(self, position=50):
        """
        Add an extra axis along the bottom of this map,
        quoting the equilibrium temperature associated
        with a particular bolometric flux received.
        """

        ax_temp = self.ax.twiny()
        mn, mx = self.ax.get_xlim()
        ax_temp.set_xscale("log")
        teqearth = (5780 * u.K / np.sqrt(2 * u.AU / u.Rsun)).to(u.K)
        ax_temp.set_xlim(teqearth * mn**0.25, teqearth * mx**0.25)
        extra_axis_color = "gray"

        ax_temp.set_xlabel(
            "Planetary Equilibrium Temperature (for zero albedo + efficient circulation)",
            color=extra_axis_color,
        )
        ax_temp.xaxis.set_ticks_position(
            "bottom"
        )  # set the position of the second x-axis to bottom
        ax_temp.xaxis.set_label_position(
            "bottom"
        )  # set the position of the second x-axis to bottom
        ax_temp.spines["bottom"].set_position(("outward", position))
        ax_temp.tick_params(axis="x", colors=extra_axis_color, which="both")
        ax_temp.spines["bottom"].set_edgecolor(extra_axis_color)
        ax_temp.xaxis.set_major_formatter(FormatStrFormatter("%dK"))
        plt.sca(self.ax)

    def plot_hz(
        self,
        Teff=[2600, 6000],
        inner="moist-greenhouse",
        outer="maximum-greenhouse",
        color="cornflowerblue",
        alpha=0.25,
        linewidth=0,
        **kw
    ):
        """
        Add a bar that indicates an approximate habitable zone.
        (Estimated very roughly by eye from Kopparapu et al.)
        """

        # make sure the plotting happens in this map
        plt.sca(self.ax)

        # define the functions
        S_inner = make_hz(inner)
        S_outer = make_hz(outer)

        if np.atleast_1d(Teff).size == 1:
            plt.axvspan(
                S_inner(Teff),
                S_outer(Teff),
                color=color,
                alpha=alpha,
                linewidth=linewidth,
                **kw
            )
        elif np.atleast_1d(Teff).size == 2:
            N = 10
            for Teff in np.linspace(Teff[0], Teff[1], N):
                plt.axvspan(
                    S_inner(Teff),
                    S_outer(Teff),
                    color=color,
                    alpha=alpha / N,
                    linewidth=0,
                    **kw
                )

        # Teff_inner = 5780
        # Teff_outer = 2600


class Flux_x_Mass(Flux_x_Radius):
    yaxis = KludgedMass


class Flux_x_Teff(BubbleMap):
    xaxis = Flux
    yaxis = StellarTeff

    def plot_hz(
        self,
        inner="moist-greenhouse",
        outer="maximum-greenhouse",
        color="cornflowerblue",
        alpha=0.25,
        linewidth=0,
        zorder=-100,
        **kw
    ):
        """
        Add a bar that indicates an approximate habitable zone.
        (Estimated very roughly by eye from Kopparapu et al.)
        """

        # make sure the plotting happens in this map
        plt.sca(self.ax)

        # define the functions
        S_inner = make_hz(inner)
        S_outer = make_hz(outer)

        # define the temperature grid
        Teff = np.linspace(2000, 7500)

        # plot the swath
        plt.fill_betweenx(
            Teff,
            S_inner(Teff),
            S_outer(Teff),
            color=color,
            alpha=alpha,
            linewidth=linewidth,
            zorder=zorder,
            **kw
        )


class SemimajorAxis_x_StellarLuminosity(BubbleMap):
    xaxis = SemimajorAxis
    yaxis = StellarLuminosity

    def plot_hz(
        self,
        inner="moist-greenhouse",
        outer="maximum-greenhouse",
        color="cornflowerblue",
        alpha=0.25,
        linewidth=0,
        zorder=-100,
        **kw
    ):
        """
        Add a bar that indicates an approximate habitable zone.
        (Estimated very roughly by eye from Kopparapu et al.)
        """

        # make sure the plotting happens in this map
        plt.sca(self.ax)

        # define the functions
        S_inner = make_hz(inner)
        S_outer = make_hz(outer)

        # define the temperature grid
        Teff = np.linspace(2000, 7500)

        # plot the swath
        plt.fill_betweenx(
            Teff,
            S_inner(Teff),
            S_outer(Teff),
            color=color,
            alpha=alpha,
            linewidth=linewidth,
            zorder=zorder,
            **kw
        )


class Distance_x_Radius(BubbleMap):
    xaxis = Distance
    yaxis = Radius


class Distance_x_Teff(BubbleMap):
    xaxis = Distance
    yaxis = StellarTeff


class EscapeParameter_x_Radius(BubbleMap):
    xaxis = EscapeParameter
    yaxis = Radius


class Density_x_Radius(BubbleMap):
    xaxis = Density
    yaxis = Radius


class StellarRadius_x_PlanetRadius(BubbleMap):
    xaxis = StellarRadius
    yaxis = Radius


class Depth_x_Radius(BubbleMap):
    xaxis = Depth
    yaxis = Radius


class Transmission_x_Radius(BubbleMap):
    xaxis = Transmission
    yaxis = Radius


class Reflection_x_Radius(BubbleMap):
    xaxis = Reflection
    yaxis = Radius


class Emission_x_Radius(BubbleMap):
    xaxis = Emission
    yaxis = Radius


class Distance_x_Brightness(BubbleMap):
    xaxis = Distance
    yaxis = StellarBrightnessTelescope


class Depth_x_Brightness(BubbleMap):
    xaxis = Depth
    yaxis = StellarBrightnessTelescope

    def plot_sigma(self, color="black", linewidth=3, alpha=0.5, **kw):
        """
        Plot the 1-sigma uncertainty from photon noise,
        with sigma on the x axis and number of photons on y-axis,
        in the telescope units associated with the map.

        Parameters
        ----------
        All keywords will be passed to plt.plot
        """

        w = self.plottable["y"].wavelength
        photons = np.logspace(0, 14) * u.ph
        sigma = 1 / np.sqrt(photons).decompose().value

        # what is "1" (e.g. JWST for one hour at R=20)?
        telescope_unit = self.plottable["y"].telescope_unit

        # how many photons do we collect with that one?
        photons_collected = photons / telescope_unit

        unit = self.plottable["y"].unit
        photons_in_unit = photons_collected.to(unit)

        plt.plot(
            sigma, photons_in_unit, color=color, linewidth=linewidth, alpha=alpha, **kw
        )


class Transmission_x_Brightness(Depth_x_Brightness):
    xaxis = Transmission


class Reflection_x_Brightness(Depth_x_Brightness):
    xaxis = Reflection


class Emission_x_Brightness(Depth_x_Brightness):
    xaxis = Emission


class Period_x_Radius(BubbleMap):
    xaxis = Period
    yaxis = Radius


class SemimajorAxis_x_Radius(BubbleMap):
    xaxis = SemimajorAxis
    yaxis = Radius


class SemimajorAxis_x_Mass(SemimajorAxis_x_Radius):
    xaxis = SemimajorAxis
    yaxis = KludgedMass


class Mass_x_Radius(ErrorMap):
    xaxis = Mass
    yaxis = Radius

    def plot_both_seager(self, **kw):
        plt.sca(self.ax)
        plot_both_seager(**kw)

    def plot_constant_density(
        self,
        densities=10.0 ** np.arange(-4, 7) * u.g / u.cm**3,
        color="coral",
        alpha=0.5,
        zorder=-100,
        **kw
    ):
        for density in densities:
            mass = np.logspace(-2, 3) * u.Mearth
            radius = (mass / density / 4 / np.pi * 3) ** (1 / 3)
            plt.plot(
                mass, radius.to(u.Rearth), color=color, alpha=alpha, zorder=zorder, **kw
            )


class Flux_x_EscapeVelocity(BubbleMap):
    xaxis = Flux
    yaxis = EscapeVelocity

    def plot_constant_lambda(
        self, alpha=0.5, color="gray", x=0.01, y=100, rotation=-4.5, **kw
    ):
        """
        Plot the escape velocity vs insolation for
        different constant values of the escape parameter.

        This assumes the exosphere is at the planet's equilibrium
        temperature, which is is *terrible* approximation. This
        also represented the *current* escape parameter, saying
        nothing about the history of XUV radiation the planet
        may have received.

        Parameters
        ----------
        **kw : dict
            Keyword parameters get passed along to plot.
        """

        # label one of the lines
        plt.text(
            x,
            y,
            r"$E_{grav}/E_{thermal}$",
            rotation=rotation,
            fontsize=8,
            color=color,
        )

        teq = np.logspace(1, 4) * u.K
        earth_teq = (5780 * u.K / np.sqrt(2 * u.AU / u.Rsun)).to(u.K)

        relative_insolation = (teq / earth_teq) ** 4
        m = 1 * u.M_p

        def escape_velocity(T, lam=1):
            # calculate the escape velocity for a particular escape parameter
            kT = con.k_B * T
            return np.sqrt(2 * kT * lam / m).to("km/s")

        plt.sca(self.ax)
        max_insolation = self.xlim[1]
        max_teq = earth_teq * max_insolation**0.25

        # loop over factors of 10 of lambda
        for lam in 10 ** np.arange(6):
            plt.plot(
                relative_insolation,
                escape_velocity(teq, lam=lam),
                color=color,
                alpha=alpha,
                **kw
            )

    def plot_jeans_shoreline(
        self,
        reference_relative_insolation=0.43337238,
        reference_escape_velocity=5.02270418 * u.km / u.s,
        alpha=0.5,
        color="gray",
        x=0.01,
        y=100,
        **kw
    ):
        relative_insolation = np.logspace(-5, 5)
        # v ~ T**(1/2) ~ I**(1/8)
        escape_velocity = (
            reference_escape_velocity
            * (relative_insolation / reference_relative_insolation) ** 0.125
        )
        plt.plot(relative_insolation, escape_velocity, alpha=alpha, color=color)

        plt.text(
            reference_relative_insolation,
            reference_escape_velocity,
            r"$v_{thermal} \propto v_{escape}$    ",
            fontsize=8,
            color=color,
            alpha=alpha,
            ha="right",
            **kw
        )

    def plot_shoreline(
        self,
        reference_relative_insolation=0.43337238,
        reference_escape_velocity=5.02270418 * u.km / u.s,
        alpha=0.5,
        color="blue",
        x=0.01,
        y=100,
        **kw
    ):
        relative_insolation = np.logspace(-5, 5)
        escape_velocity = (
            reference_escape_velocity
            * (relative_insolation / reference_relative_insolation) ** 0.25
        )
        plt.plot(relative_insolation, escape_velocity, alpha=alpha, color=color)

        plt.text(
            reference_relative_insolation,
            reference_escape_velocity,
            r"   $I \propto v_{escape}^4$",
            fontsize=8,
            color=color,
            alpha=alpha,
            **kw
        )


class EscapeVelocity_x_Flux(BubbleMap):
    xaxis = EscapeVelocity
    yaxis = Flux

    def plot_jeans_shoreline(
        self,
        reference_relative_insolation=0.43337238,
        reference_escape_velocity=5.02270418 * u.km / u.s,
        alpha=0.5,
        color="gray",
        x=0.01,
        y=100,
        **kw
    ):
        relative_insolation = np.logspace(-5, 5)
        # v ~ T**(1/2) ~ I**(1/8)
        escape_velocity = (
            reference_escape_velocity
            * (relative_insolation / reference_relative_insolation) ** 0.125
        )
        plt.plot(
            escape_velocity,
            relative_insolation,
            alpha=alpha,
            color=color,
            label=r"$v_{thermal} \propto v_{escape}$",
        )

    def plot_shoreline(
        self,
        reference_relative_insolation=0.43337238,
        reference_escape_velocity=5.02270418 * u.km / u.s,
        alpha=0.5,
        color="blue",
        x=0.01,
        y=100,
        **kw
    ):
        relative_insolation = np.logspace(-5, 5)
        escape_velocity = (
            reference_escape_velocity
            * (relative_insolation / reference_relative_insolation) ** 0.25
        )
        plt.plot(
            escape_velocity,
            relative_insolation,
            alpha=alpha,
            color=color,
            label=r"$I \propto v_{escape}^4$",
        )


class EscapeVelocity_x_CumulativeXUVFlux(EscapeVelocity_x_Flux):
    yaxis = CumulativeXUVFlux


class CumulativeXUVFlux_x_EscapeVelocity(Flux_x_EscapeVelocity):
    xaxis = CumulativeXUVFlux


class ImpactVelocity_x_EscapeVelocity(Flux_x_EscapeVelocity):
    xaxis = ImpactVelocity

    def plot_shoreline(
        self,
        reference_impact_velocity=24.650357 * u.km / u.s,
        reference_escape_velocity=5.02270418 * u.km / u.s,
        alpha=0.5,
        color="brown",
        x=0.01,
        y=100,
        rotation=-4.5,
        **kw
    ):
        impact_velocity = np.logspace(0, 3) * u.km / u.s
        escape_velocity = (
            reference_escape_velocity
            * (impact_velocity / reference_impact_velocity) ** 0.25
        )
        plt.plot(impact_velocity, escape_velocity, alpha=alpha, color=color)

        plt.text(
            reference_impact_velocity,
            reference_escape_velocity,
            r"   $v_{impact} \propto {v_{escape}}$",
            fontsize=8,
            color=color,
            alpha=alpha,
            va="center",
            ha="left",
        )


class RA_x_Dec(BubbleMap):
    xaxis = RightAscension
    yaxis = Declination

    def add_monthaxis(self, ax=None, position=40):
        """
        Add an extra axis along the bottom of this map,
        quoting the month when things are best to observe
        """
        ax_month = ax.twiny()
        ax_month.set_xticks([24, 18, 12, 6, 0], ["Sep", "Jun", "Mar", "Dec", "Sep"])
        extra_axis_color = "gray"
        ax_month.xaxis.set_ticks_position("bottom")
        ax_month.xaxis.set_label_position("bottom")
        ax_month.spines["bottom"].set_position(("outward", position))
        ax_month.tick_params(axis="x", colors=extra_axis_color, which="both")
        ax_month.spines["bottom"].set_edgecolor(extra_axis_color)


preset_maps = {}
local_variables = dict(**locals())
for k, v in local_variables.items():
    try:
        assert issubclass(v, Map)
        preset_maps[k] = v
    except (AssertionError, TypeError):
        continue
for k in ["Map", "BubbleMap", "ErrorMap"]:
    preset_maps.pop(k)



"""
Define a downloader tool to somewhat politely use a local file and/or
download a big one from online, if it's older than some particular date.
"""

from ..imports import *


class Downloader:
    """
    ...to aid downloading large archive files, storing them locally,
    and redownloading them only when/if absolutely necessary.
    """

    # class-specific keywords to help reading files
    read_kw = dict()

    def __init__(
        self,
        expiration=1.0 * u.day,
        timeout=20 * u.minute,
    ):
        """
        Initialize a Downloader.

        Parameters
        ----------
        expiration : astropy.units.quantity.Quantity
            After how long should we consider the raw downloaded table expired,
            and therefore download a new one?
        timeout : astropy.units.quantity.Quantity
            What's the longest we should wait before giving up on the download?
        """

        self._expiration = expiration
        self.timeout = timeout

    def get_table(self, remake=False):
        """
        Get the table, downloading it from online if necessary.
        If the table is older than the downloader's threshold,
        then ask the user whether or not they want to download.

        Parameters
        ----------
        remake : bool
            Whether we should (definitely) re-download the data.
                If True, download data no matter what.
                If None, download data only if (there is none) or (it's expired and the user agrees).
                If False, download data only if there is none.
        """

        # if file doesn't exist, download it

        if remake == True:
            should_we_download = True
        if remake == None:
            should_we_download = check_if_needs_updating(self.path, self._expiration)
        elif remake == False:
            should_we_download = os.path.exists(self.path) == False

        # either download a fresh file, or load a local one
        if should_we_download:
            self.download_fresh()
        else:
            print(f"Loading local file from {self.path}")

        # read the actual file
        table = Table.read(self.path, **self.read_kw)

        # possibly add some new metadata to the table
        self.add_metadata(table)

        return table

    def download_fresh(self):
        """
        Download a brand new table from the Exoplanet Archive.
        """

        print(
            f"""
        Attempting to freshly download data from
        {self.url}
        
        This may take a long time, and we're dreadully 
        sorry that we haven't coded in a clever way to 
        tell you how long. Archive tables might be up to
        a few 100MB, so however long it takes your computer
        to download that is probably a reasonable-ish amount 
        of time to wait before getting stressed out. 
        """
        )

        # download the file from the URL (20-minute timeout)
        temporary_path = download_file(
            self.url, cache=False, timeout=self.timeout.to_value("second")
        )

        # copy the file to its new location
        shutil.copyfile(temporary_path, self.path)

        print(f"Download successful! Saved file to {self.path}")

    def add_metadata(self, table):
        pass



from .population_core import *


class PredefinedPopulation(Population):
    """
    Predefined Populations are derived from archival data tables,
    representing the basic building blocks for many subsamples.
    """

    label = "?"
    _expiration = 10 * u.day

    def __init__(self, remake=False, standard=None, label=None, **plotkw):
        """
        Initialize a predefined population.

        This will first try to load a standardarized local file. If that doesn't work,
        it will download raw data  (or load a recent local cached copy) and then
        carefully create a standard table, which can be saved for future use.
        Behind the scenes, this initialization is effectively something like:

        #   ._ingest_standardized_data
        #   ...or...
        #   .._ingest_raw_data (= ._download_raw_data + ._create_standardized_data)

        Parameters
        ----------
        remake : bool
            Whether we re-ingest this table from its raw ingredients.
        standard : astropy.table.QTable
            A standardized table with which to initialize this population;
            this is mostly used for creating subsets of existing predefined
            populations without having to reinitialize a cumbersome full
            object each time.
        label : str
            A custom label to apply to this population.
        **plotkw : dict
            All other keywords are stored as plotting suggestions.
        """

        if standard is None:
            try:
                # try to load the standardized table
                assert remake == False
                standard = self._ingest_standardized_data()
            except (IOError, FileNotFoundError, AssertionError):
                # or create a new standardized table and save it
                standard = self._ingest_raw_data(remake=remake)

        assert type(standard) in [QTable, Table, Row]

        # initialize with a standard table
        Population.__init__(
            self,
            standard=standard,
            label=label or self.label,
            **plotkw,
        )

    def _ingest_raw_data(self, remake=None):
        """
        Ingest a new population table of arbitrary format,
        and then standardize it, using the tools defined in
        inherited population classes.

        Parameters
        ----------
        remake : bool
            Whether we should (definitely) re-download the data.
                If True, download data no matter what.
                If None, download data only if (there is none) or (it's expired and the user agrees).
                If False, download data only if there is none.
        """

        # load the raw table
        raw = self._download_raw_data(remake=remake)

        # create a standardized table from the array
        standard = self._create_standardized(raw)

        # save the standardized table as an ascii table for humans to read
        standard.write(
            self._standardized_data_path, format="ascii.ecsv", overwrite=True
        )
        print(f"Saved a standardized text table to {self._standardized_data_path}")

        return standard

    def _download_raw_data(self, remake=False):
        """
        Load raw data, possibly downloading them if needed.

        Parameters
        ----------
        remake : bool
            Should the raw tables be re-made/re-downloaded,
            even if recent local ones exist?

        Returns
        -------
        raw : astropy.table.QTable
            A raw, untrimmed, unstandardized table.
        """

        #
        self._raw = self._downloader.get_table(remake=remake)

        # return both, so they can both be turned into standardized tables
        return self._raw

    def _ingest_standardized_data(self):
        """
        Load a standardized population table.

        This will *try* to ingest a standardized population table from
        a local file, probably one created by `_create_standardized_table`.
        It can/will fail with either an IOError or FileNotFoundError if
        that file doesn't exist, likely prompting the `__init__()` that
        probably called this to run `._create_standardized_data()` to
        create and save a new standard table.

        Returns
        -------
        standard : astropy.table.QTable
            A table of planet properties,
            with a minimal set of columns.
        """

        # keywords for reading a standardized table
        read_kw = dict(format="ecsv", fill_values=[("", np.nan), ("--", np.nan)])

        standard = ascii.read(self._standardized_data_path, **read_kw)
        print(f"Loaded standardized table from {self._standardized_data_path}")

        return standard



from .test_buckets import *
from .test_downloaders import *
from .test_imports import *
from .test_models import *
from .test_populations import *
from .test_uncertainties import *
from .test_visualizations import *
from .test_whatsup import *
from .test_snr import *
from .test_curation import *
from .test_calculations import *



# general class for plotting exoplanet populations
# all other maps derive from this one

from ...imports import *
from ..plottables.plottable import *
from ...populations import Population


def clean_pops(initial):
    """
    Make sure the populations are stored as a dictionary,
    with reasonable labels for each as the keys.

    Parameters
    ----------
    initial : dict, list, Population
        A dictionary containing zero or more populations.
        Or, a list of one or more populations.

    Returns
    -------
    pops : dict
        A dictionary containing one or more populations.
    """

    if type(initial) == dict:
        return initial
    elif type(initial) == list:
        d = {}
        for p in initial:
            label = p.label
            if label in d:
                label += "+"
            p.label = label
            d[label] = p
        return d
    else:
        # (otherwise, assume it's already a single population)
        return {initial.label: initial}


class Map:
    # define some defaults
    title = None
    xaxis = None
    yaxis = None

    def __init__(self, xaxis=None, yaxis=None, label=None, ax=None, **kw):
        """
        Initialize a plotting map.

        Parameters
        ----------
        xaxis : Plottable, str, None
            What should the x-positions encode?
        yaxis : Plottable, str, None
            What should the y-positions encode?
        label : str
            What is a label to use for this plot,
            as a key in dictionaries and for filenames?
        ax :
            Into what ax should we place this plot?
            If None, new plotting axes will be created
            when necessary.
        """

        # set up empty populations
        self.populations = {}

        # the matplotlib objects to access the plotted points themselves
        self.scattered = {}

        # the matplotlib objects to access text labels in the plot
        self.annotated = {}

        # the 2-4 quantities being experssed in this plot
        self.plottable = {}

        # set up the x and y axes
        self.plottable["x"] = clean_plottable(xaxis or self.xaxis, **kw)
        self.plottable["y"] = clean_plottable(yaxis or self.yaxis, **kw)

        # store a label for this map
        self.label = (
            label or f'{self.plottable["x"].source}_x_{self.plottable["y"].source}'
        )

        # apply axis labels, scales, limits appropriately
        for axis in "xy":
            for attribute in ["label", "scale", "lim"]:
                setattr(
                    self, f"{axis}{attribute}", getattr(self.plottable[axis], attribute)
                )

        # assign a plotting ax to this map (maybe)
        self.ax = ax

    def __deepcopy__(self, memo):
        """
        When doing a deepcopy, copy everything except
        the `populations` dictionary, which we might
        not want to carry over.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == "populations":
                setattr(result, k, {})
            elif k == "pop":
                setattr(result, k, None)
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        return result

    def copy(self):
        """
        Create a blank copy of this Map.
        """
        new = copy.deepcopy(self)
        new.scattered = {}
        new.annotated = {}
        return new

    def __repr__(self):
        """
        How should this plottable axis be represented as a string?
        """
        individual_plottable_strings = [
            f"{k:>6} = {v}" for k, v in self.plottable.items() if v is not None
        ]
        plottable_string = "\n".join(individual_plottable_strings)
        return f"🖼️ {self.__class__.__name__}\n{plottable_string}\n"

    def point_at(self, pop):
        """
        Point this map at a particular population. This will be called
        when building up multiple populations on the same plot.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        """

        if isinstance(pop, Population):
            if pop.label not in self.populations:
                self.pop_key = pop.label
                self.populations[self.pop_key] = pop
            self.pop = self.populations[self.pop_key]

        elif pop in self.populations:
            # if a key, focus on Population from dictionary
            self.pop = self.populations[pop]
            self.pop_key = pop
        else:
            raise ValueError(
                f"""
            It's not clear how to interpret {pop}
            as a population at which we might point 
            {self}"""
            )

    @property
    def x(self):
        return self.plottable["x"].value(self.pop)

    @property
    def y(self):
        return self.plottable["y"].value(self.pop)

    @property
    def x_uncertainty(self):
        return self.plottable["x"].uncertainty(self.pop)

    @property
    def y_uncertainty(self):
        return self.plottable["y"].uncertainty(self.pop)

    @property
    def x_uncertainty_lowerupper(self):
        return self.plottable["x"].uncertainty_lowerupper(self.pop)

    @property
    def y_uncertainty_lowerupper(self):
        return self.plottable["y"].uncertainty_lowerupper(self.pop)

    def setup_axes(self, ax=None):
        """
        Set up the axes for this Map.

        Parameters
        ----------
        ax :
            The axes into which this map should be placed.
            If no axes are provided, a new figure + axes will be created
        """

        self.ax = ax or self.ax
        if self.ax is None:
            fi, self.ax = plt.subplots(1, 1, constrained_layout=True)
        assert self.ax is not None
        plt.sca(self.ax)

        """# make sure we're pointing at the axes for this map
        if ax is not None:
            # if an ax is provided, point at that
            self.ax = ax

        # if need be, create a new axes for this map
        try:
            self.ax != None
        except (AttributeError, AssertionError):
            fi, self.ax = plt.subplots(1, 1, constrained_layout=True)

        assert self.ax is not None"""

    def kw(self, pop=None, **kw):
        """
        Do a little decision-making about the plotting keyword
        arguments, pulling defaults from each population where
        needed.

        Parameter
        ---------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        **kw : dict
            All other keywords will be directed toward
            overwriting individual population defaults.
        """

        # make sure we're pointing at the right population
        if pop != None:
            self.point_at(pop)

        # define some default keywords, which can be over-written
        default = dict(
            alpha=self.pop.alpha,
            zorder=self.pop.zorder,
            label=self.pop.label,
        )

        # if any other keywords are provided, overwrite these defaults
        default.update(**kw)

        return default

    def plot(self, pop, ax=None, annotate_kw={}, **kw):
        """
        Add the points for one population to this map.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        ax :
            Into what ax should we place this plot?
            If None, create a new plotting axes.
        annotate_kw : dict
            Keywords for labeling the planet names.
        **kw : dict
            Any extra keywords will be passed on to `scatter`
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # focus attention on that population
            self.point_at(pop)

            # make sure we're plotting into the appropriate axes
            self.setup_axes(ax=ax)

            # add the scattered points

            these_scattered_points = self.ax.scatter(self.x, self.y, **self.kw(**kw))

            if self.pop_key in self.scattered:
                warnings.warn(
                    f"""
                Key '{self.pop_key}' already exists. This might be fine for plotting, 
                but if you want to access any of the plotted elements to modify them 
                later, you might not be able to because they may have been overwritten.
                Might we please encourage you to give your populations unique labels 
                via the `population.label = "here's some neat label"`?
                """
                )
            self.scattered[self.pop_key] = these_scattered_points

            # set the scales, limits, labels
            self.refine_axes()

            # add planet or hostname labels
            self.add_system_annotations(annotate_kw=annotate_kw)

    def refine_axes(self):
        """
        Update the basic axes properties after plotting.

        This applies the default scale, limits, and axis
        labels to this Map, after data have been plotted
        in it. Technically, it only needs to be called after
        all populations have been added to a plot, but
        we call it at the end of each `.plot()` to make
        sure it happens at least once.
        """
        # implement the default scales, limiets, labels for the axes
        self.ax.set_xscale(self.xscale)
        self.ax.set_yscale(self.yscale)
        self.ax.set_ylim(*self.ylim)
        self.ax.set_xlim(*self.xlim)
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)

    def build(
        self,
        pops={},
        save=False,
        steps=True,
        format="png",
        savefig_kw={},
        legend=False,
        legend_kw={},
        **kw,
    ):
        """
        Build up this map by plotting every population into it.

        Parameters
        ----------
        pops : dict
            A dictionary of populations, with keys as labels values as
            initialized populations. For example,

                pops = {'solarsystem':SolarSystem(),
                        'transiting':TransitingExoplanets()}
        save : bool
            Should we save the figure(s) to files?
        steps : bool
            Should we save intermediate steps,
            with subsets of populations, to files?
        format : str
            What kind of graphics file should be saved?
            Any format allowed by `plt.savefig` is fine;
            we recommend `png` or `pdf`.
        savefig_kw : dict
            Keywords to pass to `plt.savefig` for saving figures.
        legend : bool
            Should we include a legend?
        legend_kw : dict
            Keywords to pass to legend.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # start with a dictionary of populations
            self.populations = clean_pops(pops)

            # set up figure saving
            if save:
                filename = f"{self.label}"
                directory = filename
                mkdir(directory)

            # set up legend defaults
            lkw = dict(frameon=False)
            lkw.update(**legend_kw)

            # loop over all the populations
            for key in self.populations.keys():

                # plot each population, passing plotting keywords to it
                self.plot(key, **kw)

                # make a legend, if requested
                if legend:
                    self.add_legend(**lkw)

                # update filename and save intermediate figures
                if save:
                    filename += f"+{key}"
                    if steps:
                        plt.savefig(f"{directory}/{filename}.{format}", **savefig_kw)

            # save the final figure
            if save and not steps:
                plt.savefig(f"{directory}/{filename}.{format}", **savefig_kw)

    def add_system_annotations(self, annotate_kw={}):
        """
        Add text annotations for planet names.
        FIXME!
        """
        # if requested, label the individual planets in the plot
        if self.pop.annotate_planets:
            kw = dict(**annotate_kw)
            if getattr(self.pop, "zorder", None) is not None:
                kw["zorder"] = self.pop.zorder
            kw.update(**getattr(self.pop, "annotate_kw", {}))
            self.annotate_planets(self.pop, **kw)

    def annotate_planets(self, pop, before="\n", after="", restrictlimits=False, **kw):
        """
        Label the planets in whatever population we're pointed at.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        before : str
            Add the string at the start of each name.
        after : str
            Add the string at the end of each name.
        restrictlimits : bool
            Should we plot names only for those planets that fall within
            the map's x and y limits?
        **kw : dict
            Any additional keywords will be passed to the `text` command.
        """

        # make sure we're pointing at this population
        self.point_at(pop)

        # make sure we're set to the current axes
        plt.sca(self.ax)

        # loop over the elements in the population
        for i in range(len(self.x)):
            # pull out the positions and the name
            x, y, name = self.x[i], self.y[i], self.pop.name()[i]

            # skip over the planets that aren't within limits
            if restrictlimits:
                try:
                    with np.errstate(invalid="ignore"):
                        if x < np.min(self.xlim) or x > np.max(self.xlim):
                            if y < np.min(self.ylim) or y > np.max(self.ylim):
                                continue
                except TypeError:  # (are there other errors we need to add?)
                    pass

            # define some defaults for the text
            textkw = dict(
                ha="center",
                va="top",
                fontsize=6,
                color=self.pop.color,
                alpha=self.pop.alpha,
                clip_on=True,
            )

            # think this is just as Python 3 thing
            textkw.update(**kw)

            # store the text plot, so it can be modified
            try:
                assert np.isfinite(x * y)
                self.annotated[name] = plt.text(x, y, before + name + after, **textkw)
            except AssertionError:  # (do we need to add other errors?)
                pass

    def annotate_hosts(
        self, pop, before="\n", after="", restrictlimits=False, once=False, **kw
    ):
        """
        Label the planet hosts in whatever population we're pointed at.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        before : str
            Add the string at the start of each name.
        after : str
            Add the string at the end of each name.
        restrictlimits : bool
            Should we plot names only for those planets that fall within
            the map's x and y limits?
        **kw : dict
            Any additional keywords will be passed to the `text` command.
        """

        # make sure we're pointing at this population
        self.point_at(pop)

        # make sure we're set to the current axes
        plt.sca(self.ax)

        # loop over the elements in the population
        for i in range(len(self.x)):
            # pull out the positions and the name
            x, y, name = self.x[i], self.y[i], self.pop.hostname()[i]

            if once:
                if name in self.annotated:
                    continue

            # skip over the planets that aren't within limits
            if restrictlimits:
                try:
                    with np.errstate(invalid="ignore"):
                        if x < np.min(self.xlim) or x > np.max(self.xlim):
                            if y < np.min(self.ylim) or y > np.max(self.ylim):
                                continue
                except TypeError:  # (are there other errors we need to add?)
                    pass

            # define some defaults for the text
            textkw = dict(
                ha="center",
                va="top",
                fontsize=6,
                color=self.pop.color,
                alpha=self.pop.alpha,
                clip_on=True,
            )

            # think this is just as Python 3 thing
            textkw.update(**kw)

            # store the text plot, so it can be modified
            try:
                assert np.isfinite(x * y)
                self.annotated[name] = plt.text(x, y, before + name + after, **textkw)

            except AssertionError:  # (do we need to add other errors?)
                pass

    def connect_planets(self, pop, **kw):
        """
        Identify all the multiplanet systems,
        and draw lines connecting the planets
        within each individual system.

        Parameters
        ----------
        pop : Population, str, int
            The population to plot. This could be either an
            actual Population object, or a key referring to
            an element of the `self.populations` dictionary.
        **kw : dict
            Any keywords will overwrite the defaults
            going into the `.plot()` function call.
        """

        # make sure we're pointing at this population
        self.point_at(pop)

        # make sure we're set to the current axes
        plt.sca(self.ax)

        # define some defaults for the lines
        linekw = dict(color=self.pop.color, alpha=self.pop.alpha * 0.5, zorder=-100)

        # overwrite those defaults if kw are provided
        linekw.update(**kw)

        original_pop = self.pop
        for hostname in np.unique(self.pop.hostname()):
            friends = self.pop[hostname]
            self.pop = friends
            x, y = self.x, self.y
            i = np.argsort(x)
            plt.plot(x[i], y[i], **linekw)
            self.pop = original_pop

    def remove_xlabel(self):
        """
        Remove the xlabel and xticklabels from this map.
        """
        plt.setp(self.ax.get_xticklabels(), visible=False)
        self.ax.set_xlabel("")

    def remove_ylabel(self):
        """
        Remove the ylabel and xticklabels from this map.
        """
        plt.setp(self.ax.get_yticklabels(), visible=False)
        self.ax.set_ylabel("")

    def ticks_simplify_exponents(self, which="xy"):
        if "x" in which:
            self.ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))
        if "y" in which:
            self.ax.yaxis.set_major_formatter(FormatStrFormatter("%d"))

    def ticks_enforce_multiple_oom(self, which="xy"):
        if "x" in which:
            loc = LogLocator(base=10, numticks=3)
            self.ax.xaxis.set_major_locator(loc)
        if "y" in which:
            loc = LogLocator(base=10, numticks=3)
            self.ax.yaxis.set_major_locator(loc)

    def add_legend(self, outside=False, frameon=False, **kw):
        plt.sca(self.ax)
        legendkw = dict()
        legendkw["frameon"] = frameon
        if outside:
            legendkw["bbox_to_anchor"] = (1, 1)
            legendkw["loc"] = "upper left"
        legendkw.update(**kw)
        legend = plt.legend(**legendkw)
        legend.set_in_layout(False)



from .setup_tests import *

from exoatlas.imports import *
from exoatlas.populations import TransitingExoplanets
from exoatlas.whatsup import Plan


def test_whatsup():
    with mock.patch("builtins.input", return_value=""):
        planets = TransitingExoplanets()

    p = Plan(
        planets,  # what should we try?
        start=Time("2020-03-14"),  # when should we start?
        finish=Time("2020-03-14"),  # when should we stop?
        observatory="SBO",  # what observatory are starting from?
        directory="upcoming-transits",
    )  # where should we


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



from ...imports import *
from ..predefined import PredefinedPopulation

__all__ = ["SolarSystem", "SolarSystemDwarfPlanets"]


class SolarSystem(PredefinedPopulation):
    """The Solar System, very crudely."""

    _raw_mass_unit = u.Unit(1e24 * u.kg)
    _filename = "jpl-ssd-planets.txt"

    label = "Solar System"
    # the data in the table probably don't need to be updated
    _expiration = np.inf * u.day

    def __init__(self, **kwargs):
        """
        Initialize a population of Solar System (main) planets.
        """
        PredefinedPopulation.__init__(self, **kwargs)
        self.color = "cornflowerblue"
        self.zorder = 1e10
        self.s = 80
        self.respond_to_color = False
        self.exact = True
        self.marker = "s"

    def _download_raw_data(self, **kw):
        """
        Load the raw table of data for the Solar System.
        """

        # load the table of Solar System planets
        initial_filename = os.path.join(
            code_directory, f"populations/data/solarsystem/{self._filename}"
        )
        raw = ascii.read(initial_filename)

        # for debugging, hang on to the raw table as a hidden attribute
        self._raw = raw

        # a table of unstandardized planet properties
        return raw

    def _create_standardized(self, trimmed):
        """
        Create a standardized table of planet properties.
        """

        # start from the trimmed table
        t = QTable(trimmed)

        # create an empty standardized table
        s = QTable()
        s["name"] = t["name"]
        s["hostname"] = "Sun"

        # set up the Sun
        s["stellar_teff"] = 5780 * u.K
        s["stellar_radius"] = 1.0 * u.Rsun
        s["stellar_mass"] = 1.0 * u.Msun
        s["stellar_luminosity"] = 1 * u.Lsun
        s["stellar_age"] = 4.57 * u.Gyr

        # store the period, by default, in day
        s["period"] = (t["period"] * u.year).to(u.day)

        def add_with_uncertainty(
            their_name, our_name, their_unit=u.Unit(), our_unit=u.Unit()
        ):
            s[our_name] = (t[their_name] * their_unit).to(our_unit)
            try:
                s[f"{our_name}_uncertainty_{k}"] = (
                    t[f"{their_name}_uncertainty"] * their_unit
                ).to(our_unit)
            except KeyError:
                pass

        add_with_uncertainty("equatorial_radius", "radius_equatorial", u.km, u.Rearth)
        add_with_uncertainty("mean_radius", "radius", u.km, u.Rearth)
        add_with_uncertainty("mass", "mass", self._raw_mass_unit, u.Mearth)
        add_with_uncertainty("density", "density", u.g / u.cm**3, u.g / u.cm**3)
        add_with_uncertainty("rotational_period", "rotational_period", u.day, u.day)
        add_with_uncertainty("period", "period", u.year, u.day)
        add_with_uncertainty("absolute_magnitude", "absolute_magnitude", u.mag, u.mag)
        add_with_uncertainty("albedo_geometric", "albedo_geometric")
        add_with_uncertainty(
            "surface_gravity", "surface_gravity", u.m / u.s**2, u.m / u.s**2
        )
        add_with_uncertainty(
            "escape_velocity", "escape_velocity", u.km / u.s, u.km / u.s
        )
        try:
            add_with_uncertainty("eccentricity", "eccentricity")
            add_with_uncertainty(
                "argument_of_periastron", "argument_of_periastron", u.deg, u.deg
            )
            s["inclination"] = (90 - t["inclination"]) * u.deg

        except KeyError:
            s["eccentricity"] = np.nan
            s["argument_of_periastron"] = np.nan * u.deg
            s["inclination"] = 90 * u.deg

        # use Kepler's (actual) 3rd Law to get semimajor axis
        semimajoraxis = (s["period"].to(u.year).value) ** (2.0 / 3.0) * u.AU
        s["semimajoraxis"] = semimajoraxis
        # s["transit_scaled_semimajoraxis"] = (semimajoraxis / u.Rsun).decompose().value

        # equilibrium temperature assuming uniform redistribution + 0 albedo
        # s["teq"] = s["stellar_teff"] * (0.25 * 1.0 / s["transit_scaled_semimajoraxis"] ** 2) ** 0.25

        # some other planet parameters we might not need for the solar system
        # s["rv_semiamplitude"] = np.nan * u.m / u.s
        s["radius_ratio"] = (s["radius"] / s["stellar_radius"]).decompose().value
        s["distance"] = np.nan * u.pc
        s["ra"] = np.nan * 0.0 * u.deg
        s["dec"] = np.nan * 0.0 * u.deg
        # s["discovery_facility"] = "humans"
        # s["transit_midpoint"] = np.nan * u.day
        # s["transit_duration"] = np.nan * u.day
        # s["transit_depth"] = (s["radius"] / s["stellar_radius"]).decompose() ** 2
        # s["transit_impact_parameter"] = 0.0
        # s["inclination"] = 90 * u.deg

        s

        self.standard = s
        return s


class SolarSystemDwarfPlanets(SolarSystem):
    _raw_mass_unit = u.Unit(1e18 * u.kg)
    _filename = "jpl-ssd-dwarf-planets.txt"

    label = "Solar System Dwarf Planets"

    def __init__(self, **kwargs):
        """
        Initialize a population of Solar System (dwarf) planets.
        """
        PredefinedPopulation.__init__(self, **kwargs)
        self.color = "royalblue"
        self.s = 40
        self.respond_to_color = False
        self.exact = True
        self.marker = "s"



from ...imports import *
from ..predefined import PredefinedPopulation
from .sbdb_downloader import *


class SolarSystemMinorPlanets(PredefinedPopulation):
    label = "Solar System Minor Planets"

    def __init__(self, minimum_diameter=100 * u.km, **kw):
        self.minimum_diameter = minimum_diameter
        self._downloader = SBDBDownloader(minimum_diameter=minimum_diameter)
        PredefinedPopulation.__init__(self, **kw)

        self.color = "midnightblue"
        self.zorder = 1
        self.s = 20
        self.respond_to_color = False
        self.exact = True
        self.marker = "s"

    @property
    def fileprefix(self):
        """
        Define a fileprefix for this population, to be used
        for setting the filename of the standardized population.
        """
        return f'{clean(self.label)}-{self.minimum_diameter.to_value("km")}km'

    def _create_standardized(self, raw):
        """
        Create a standardized table to make sure that at
        least a few necessary columns are populated.

        This standardization step aims to make it easier to
        compare exoplanet populations from different sources.
        It translates columns from a raw table into a standard
        set of expected column names, doing some calculations
        if necessary.

        Parameters
        ----------
        raw : astropy.table.QTable
            A raw unstandardized table from `SBDBDownloader`.

        Returns
        -------
        standard : astropy.table.QTable
            A standardize table of exoplanet properties.
        """

        # define the table from which we're deriving everying
        r = raw

        # the new standardized table
        s = QTable()

        # names
        for k in ["name", "full_name", "class"]:
            s[k] = raw[k]

        s["eccentricity"] = raw["e"]
        s["semimajoraxis"] = raw["a"] * u.AU
        s["perihelion"] = raw["q"] * u.AU
        s["inclination"] = raw["i"] * u.deg
        s["Omega"] = raw["om"] * u.deg
        s["argument_of_periastron"] = raw["w"] * u.deg
        s["mean_anomaly"] = raw["ma"] * u.deg
        s["time_of_perihelion"] = raw["tp"] * u.day
        s["period"] = raw["per"] * u.day
        s["H"] = raw["H"]
        s["G"] = raw["G"]
        s["radius"] = (0.5 * raw["diameter"] * u.km).to(u.Rearth)
        s["mass"] = (raw["GM"] * u.km**3 / u.s**2 / con.G).to(u.Mearth)
        s["density"] = raw["density"] * u.g / u.cm**3
        s["rotational_period"] = raw["rot_per"] * u.hour
        s["albedo"] = raw["albedo"]
        s["hostname"] = "Sun"
        s["stellar_teff"] = 5780 * u.K
        s["stellar_radius"] = 1 * u.Rsun
        s["stellar_mass"] = 1 * u.Msun
        s["stellar_age"] = 4.5 * u.Gyr

        # for k in ["mass", "radius"]:
        #    for side in ["lower", "upper"]:
        #        s[f"{k}_uncertainty_{side}"] = 0 * s[k].unit

        # sort these minor planets by their names (with numbers)
        s.sort("full_name")

        # set the fill_value for any numeric columns to nans
        for k in s.colnames:
            if not isinstance(s[k][0], str):
                s[k].fill_value = np.nan

        # fill in all the masked elements to make an unmasked array with nans
        standard = s.filled()

        return standard
